#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from oslo_log import log as logging

from vif_plug_ovs import constants
from vif_plug_ovs import linux_net
from vif_plug_ovs.ovsdb import api as ovsdb_api


LOG = logging.getLogger(__name__)
QOS_UUID_NAMESPACE = uuid.UUID("68da264a-847f-42a8-8ab0-5e774aee3d95")


class BaseOVS(object):

    def __init__(self, config):
        self.timeout = config.ovs_vsctl_timeout
        self.connection = config.ovsdb_connection
        self.interface = config.ovsdb_interface
        self._ovsdb = None

    # NOTE(sean-k-mooney): when using the native ovsdb bindings
    # creating an instance of the ovsdb api connects to the ovsdb
    # to initialize the library based on the schema version
    # of the ovsdb. To avoid that we lazy load the ovsdb
    # instance the first time we need it via a property.
    @property
    def ovsdb(self):
        if not self._ovsdb:
            self._ovsdb = ovsdb_api.get_instance(self)
        return self._ovsdb

    def _ovs_supports_mtu_requests(self):
        return self.ovsdb.has_table_column('Interface', 'mtu_request')

    def _set_mtu_request(self, txn, dev, mtu):
        txn.add(
            self.ovsdb.db_set(
                'Interface', dev, ('mtu_request', mtu)
            )
        )

    def update_device_mtu(self, txn, dev, mtu, interface_type=None):
        if not mtu:
            return
        if interface_type not in [
            constants.OVS_VHOSTUSER_INTERFACE_TYPE,
            constants.OVS_VHOSTUSER_CLIENT_INTERFACE_TYPE]:
            linux_net.set_device_mtu(dev, mtu)
        elif self._ovs_supports_mtu_requests():
            self._set_mtu_request(txn, dev, mtu)
        else:
            LOG.debug("MTU not set on %(interface_name)s interface "
                      "of type %(interface_type)s.",
                      {'interface_name': dev,
                       'interface_type': interface_type})

    def ensure_ovs_bridge(self, bridge, datapath_type):
        return self.ovsdb.add_br(bridge, may_exist=True,
                                 datapath_type=datapath_type).execute()

    def delete_ovs_bridge(self, bridge):
        """Delete ovs bridge by name

        :param bridge: bridge name as a string

        .. note:: Do Not call with br-int !!!
        """
        # TODO(sean-k-mooney): when we fix bug: #1914886
        # add a guard against deleting the integration bridge
        # after adding a config option to store its name.
        return self.ovsdb.del_br(bridge).execute()

    def create_patch_port_pair(
        self, port_bridge, port_bridge_port, int_bridge, int_bridge_port,
        iface_id, mac, instance_id, tag=None
    ):
        """Create a patch port pair between any two bridges.

        :param port_bridge: the source bridge name for the patch port pair.
        :param port_bridge_port: the name of the patch port on the
        source bridge.
        :param int_bridge: the target bridge name, typically br-int.
        :param int_bridge_port: the name of the patch port on the
        target bridge.
        :param iface_id: neutron port ID.
        :param mac: port MAC.
        :param instance_id: instance uuid.
        :param mtu: port MTU.
        :param tag: OVS interface tag used for vlan isolation.
        """

        # NOTE(sean-k-mooney): we use a transaction here for 2 reasons:
        # 1.) if using the vsctl client its faster
        # 2.) in all cases we either want to fully create the patch port
        # pair or not create it atomically. By using a transaction we know
        # that we will never be in a mixed state where it was partly created.
        with self.ovsdb.transaction() as txn:
            # create integration bridge patch peer
            external_ids = {
                'iface-id': iface_id, 'iface-status': 'active',
                'attached-mac': mac, 'vm-uuid': instance_id
            }
            col_values = [
                ('external_ids', external_ids),
                ('type', 'patch'),
                ('options', {'peer': port_bridge_port})
            ]

            txn.add(self.ovsdb.add_port(int_bridge, int_bridge_port))
            if tag:
                txn.add(
                    self.ovsdb.db_set('Port', int_bridge_port, ('tag', tag)))
            txn.add(
                self.ovsdb.db_set('Interface', int_bridge_port, *col_values))

            # create port bridge patch peer
            col_values = [
                ('type', 'patch'),
                ('options', {'peer': int_bridge_port})
            ]
            txn.add(self.ovsdb.add_port(port_bridge, port_bridge_port))
            txn.add(
                self.ovsdb.db_set('Interface', port_bridge_port, *col_values))

    def create_ovs_vif_port(
        self, bridge, dev, iface_id, mac, instance_id,
        mtu=None, interface_type=None, vhost_server_path=None,
        tag=None, pf_pci=None, vf_num=None, set_ids=True, datapath_type=None,
        qos_type=None, vlan_mode=None, trunks=None
    ):
        """Create OVS port

        :param bridge: bridge name to create the port on.
        :param dev: port name.
        :param iface_id: port ID.
        :param mac: port MAC.
        :param instance_id: VM ID on which the port is attached to.
        :param mtu: port MTU.
        :param interface_type: OVS interface type.
        :param vhost_server_path: path to socket file of vhost server.
        :param tag: OVS interface tag.
        :param pf_pci: PCI address of PF for dpdk representor port.
        :param vf_num: VF number of PF for dpdk representor port.
        :param set_ids: set external ids on port (bool).
        :param datapath_type: datapath type for port's bridge
        :param qos_type: qos type for a port

        .. note:: create DPDK representor port by setting all three values:
            `interface_type`, `pf_pci` and `vf_num`. if interface type is
            not `OVS_DPDK_INTERFACE_TYPE` then `pf_pci` and `vf_num` values
            are ignored.
        """
        external_ids = {'iface-id': iface_id,
                        'iface-status': 'active',
                        'attached-mac': mac,
                        'vm-uuid': instance_id}
        col_values = [('external_ids', external_ids)] if set_ids else []
        if interface_type:
            col_values.append(('type', interface_type))
        if vhost_server_path:
            col_values.append(('options',
                               {'vhost-server-path': vhost_server_path}))
        if (interface_type == constants.OVS_DPDK_INTERFACE_TYPE and
                pf_pci and vf_num):
            devargs_string = "{PF_PCI},representor=[{VF_NUM}]".format(
                PF_PCI=pf_pci, VF_NUM=vf_num)
            col_values.append(('options',
                              {'dpdk-devargs': devargs_string}))
        # create qos record if qos type is specified
        # and get the qos id. This is done outside of the transaction
        # because we need the qos id to set the qos on the port.
        # The qos uuid cannot be set when creating the record so we
        # have to look it up after the record is created. this means
        # we need to create the qos record outside of the transaction
        # that creates the port.
        qid = None
        if qos_type:
            self.delete_qos_if_exists(dev, qos_type)
            qos_id = uuid.uuid5(QOS_UUID_NAMESPACE, dev)
            qos_external_ids = {'id': str(qos_id), '_type': qos_type}
            self.ovsdb.db_create(
                'QoS', type=qos_type, external_ids=qos_external_ids
                ).execute(check_error=True)
            record = self.get_qos(dev, qos_type)
            qid = record[0]['_uuid']

        with self.ovsdb.transaction() as txn:
            if datapath_type:
                txn.add(self.ovsdb.add_br(bridge, may_exist=True,
                                          datapath_type=datapath_type))
            txn.add(self.ovsdb.add_port(bridge, dev))
            if tag:
                txn.add(self.ovsdb.db_set('Port', dev, ('tag', tag)))
            if vlan_mode:
                txn.add(self.ovsdb.db_set('Port', dev,
                                          ('vlan_mode', vlan_mode)))
            if trunks:
                txn.add(self.ovsdb.db_set('Port', dev, ('trunks', trunks)))
            if qid:
                txn.add(self.ovsdb.db_set('Port', dev, ('qos', qid)))
            if col_values:
                txn.add(self.ovsdb.db_set('Interface', dev, *col_values))
            self.update_device_mtu(
                txn, dev, mtu, interface_type=interface_type
            )

    def port_exists(self, port_name, bridge):
        ports = self.ovsdb.list_ports(bridge).execute()
        return ports is not None and port_name in ports

    def get_qos(self, dev, qos_type):
        qos_id = uuid.uuid5(QOS_UUID_NAMESPACE, dev)
        external_ids = {'id': str(qos_id), '_type': qos_type}
        return self.ovsdb.db_find(
            'QoS', ('external_ids', '=', external_ids),
            colmuns=['_uuid']
        ).execute()

    def delete_qos_if_exists(self, dev, qos_type):
        qos_ids = self.get_qos(dev, qos_type)
        if qos_ids is not None and len(qos_ids) > 0:
            for qos_id in qos_ids:
                if '_uuid' in qos_id:
                    self.ovsdb.db_destroy(
                        'QoS', str(qos_id['_uuid'])
                    ).execute()

    def update_ovs_vif_port(self, dev, mtu=None, interface_type=None):
        with self.ovsdb.transaction() as txn:
            self.update_device_mtu(
                txn, dev, mtu, interface_type=interface_type
            )

    def delete_ovs_vif_port(
            self, bridge, dev, delete_netdev=True, qos_type=None
    ):
        self.ovsdb.del_port(dev, bridge=bridge, if_exists=True).execute()
        if qos_type:
            self.delete_qos_if_exists(dev, qos_type)
        if delete_netdev:
            linux_net.delete_net_dev(dev)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import functools
import os

from os_vif.tests.functional import base as os_vif_base


wait_until_true = os_vif_base.wait_until_true


class VifPlugOvsBaseFunctionalTestCase(os_vif_base.BaseFunctionalTestCase):
    """Base class for vif_plug_ovs functional tests."""

    COMPONENT_NAME = 'vif_plug_ovs'
    PRIVILEGED_GROUP = 'vif_plug_ovs_privileged'

    def _check_bridge(self, name):
        return self._ovsdb.br_exists(name).execute()

    def _check_port(self, name, bridge):
        return self.ovs.port_exists(name, bridge)

    @functools.cache
    def _get_timeout(self):
        return int(os.environ.get('OS_VIF_CHECK_PARAMETER_TIMEOUT', '10'))

    def _check_parameter(self, table, port, parameter, expected_value):
        def get_value():
            return self._ovsdb.db_get(table, port, parameter).execute()

        def check_value():
            val = get_value()
            return val == expected_value
        self.assertTrue(
            wait_until_true(
                check_value, timeout=self._get_timeout(), sleep=0.5),
            f"Parameter {parameter} of {table} {port} is {get_value()} "
            f"not {expected_value}"
        )

    def _add_bridge(self, name, may_exist=True, datapath_type=None):
        self._ovsdb.add_br(name, may_exist=may_exist,
                           datapath_type=datapath_type).execute()
        self.assertTrue(self._check_bridge(name))

    def _del_bridge(self, name):
        self._ovsdb.del_br(name).execute()



# Derived from nova/network/linux_net.py
#
# Copyright (c) 2011 X.commerce, a business unit of eBay Inc.
# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Implements vlans, bridges, and iptables rules using linux utilities."""

import os

from os_vif.internal.ip.api import ip as ip_lib
from oslo_concurrency import lockutils
from oslo_concurrency import processutils
from oslo_log import log as logging

from vif_plug_linux_bridge import privsep


LOG = logging.getLogger(__name__)
_IPTABLES_MANAGER = None


def _set_device_mtu(dev, mtu):
    """Set the device MTU."""
    if mtu:
        ip_lib.set(dev, mtu=mtu, check_exit_code=[0, 2, 254])
    else:
        LOG.debug("MTU not set on %(interface_name)s interface",
                  {'interface_name': dev})


def _ip_bridge_cmd(action, params, device):
    """Build commands to add/del ips to bridges/devices."""
    cmd = ['ip', 'addr', action]
    cmd.extend(params)
    cmd.extend(['dev', device])
    return cmd


@privsep.vif_plug.entrypoint
def ensure_vlan_bridge(vlan_num, bridge, bridge_interface,
                       net_attrs=None, mac_address=None,
                       mtu=None):
    """Create a vlan and bridge unless they already exist."""
    interface = _ensure_vlan_privileged(vlan_num, bridge_interface,
                                        mac_address, mtu=mtu)
    _ensure_bridge_privileged(bridge, interface, net_attrs)
    _ensure_bridge_filtering(bridge, None)
    return interface


@lockutils.synchronized('nova-lock_vlan', external=True)
def _ensure_vlan_privileged(vlan_num, bridge_interface, mac_address, mtu):
    """Create a vlan unless it already exists.

    This assumes the caller is already annotated to run
    with elevated privileges.
    """
    interface = 'vlan%s' % vlan_num
    if not ip_lib.exists(interface):
        LOG.debug('Starting VLAN interface %s', interface)
        ip_lib.add(interface, 'vlan', link=bridge_interface,
                   vlan_id=vlan_num, check_exit_code=[0, 2, 254])
        # (danwent) the bridge will inherit this address, so we want to
        # make sure it is the value set from the NetworkManager
        if mac_address:
            ip_lib.set(interface, address=mac_address,
                       check_exit_code=[0, 2, 254])
        ip_lib.set(interface, state='up', check_exit_code=[0, 2, 254])
        # NOTE(vish): set mtu every time to ensure that changes to mtu get
        #             propagated
        _set_device_mtu(interface, mtu)

    return interface


@lockutils.synchronized('nova-lock_bridge', external=True)
def ensure_bridge(bridge, interface, net_attrs=None, gateway=True,
                  filtering=True, mtu=None):
    _ensure_bridge_privileged(bridge, interface, net_attrs, gateway,
                              filtering=filtering, mtu=mtu)
    if filtering:
        _ensure_bridge_filtering(bridge, gateway)


# TODO(sean-k-mooney): extract into common module
def _disable_ipv6(bridge):
    """disable ipv6 for bridge if available, must be called from
       privsep context.
    :param bridge: string bridge name
    """
    # NOTE(sean-k-mooney): os-vif disables ipv6 to ensure the Bridge
    # does not aquire an ipv6 auto config or link local adress.
    # This is required to prevent bug 1302080.
    # https://bugs.launchpad.net/neutron/+bug/1302080
    disv6 = ('/proc/sys/net/ipv6/conf/%s/disable_ipv6' % bridge)
    if os.path.exists(disv6):
        with open(disv6, 'w') as f:
            f.write('1')


# TODO(ralonsoh): extract into common module
def _arp_filtering(bridge):
    """Prevent the bridge from replying to ARP messages with machine local IPs

    1. Reply only if the target IP address is local address configured on the
       incoming interface.
    2. Always use the best local address.
    """
    arp_params = [('/proc/sys/net/ipv4/conf/%s/arp_ignore' % bridge, '1'),
                  ('/proc/sys/net/ipv4/conf/%s/arp_announce' % bridge, '2')]
    for parameter, value in arp_params:
        if os.path.exists(parameter):
            with open(parameter, 'w') as f:
                f.write(value)


def _update_bridge_routes(interface, bridge):
    """Updates routing table for a given bridge and interface.
    :param interface: string interface name
    :param bridge: string bridge name
    """
    # TODO(sean-k-mooney): investigate deleting all this route
    # handling code. The vm tap devices should never have an ip,
    # this is old nova networks code and i dont think it will ever
    # be needed in os-vif.
    # NOTE(vish): This will break if there is already an ip on the
    #             interface, so we move any ips to the bridge
    # NOTE(danms): We also need to copy routes to the bridge so as
    #              not to break existing connectivity on the interface
    old_routes = []
    out, _ = processutils.execute('ip', 'route', 'show', 'dev',
                                  interface)
    for line in out.split('\n'):
        fields = line.split()
        if fields and 'via' in fields:
            old_routes.append(fields)
            processutils.execute('ip', 'route', 'del', *fields)

    out, _ = processutils.execute('ip', 'addr', 'show', 'dev',
                                  interface, 'scope', 'global')
    for line in out.split('\n'):
        fields = line.split()
        if fields and fields[0] == 'inet':
            if fields[-2] in ('secondary', 'dynamic', ):
                params = fields[1:-2]
            else:
                params = fields[1:-1]
                processutils.execute(*_ip_bridge_cmd('del', params,
                                                     fields[-1]),
                                     check_exit_code=[0, 2, 254])
                processutils.execute(*_ip_bridge_cmd('add', params,
                                                     bridge),
                                     check_exit_code=[0, 2, 254])
    for fields in old_routes:
        processutils.execute('ip', 'route', 'add', *fields)


@privsep.vif_plug.entrypoint
def _ensure_bridge_privileged(bridge, interface, net_attrs, gateway,
                              filtering=True, mtu=None):
    """Create a bridge unless it already exists.

    :param interface: the interface to create the bridge on.
    :param net_attrs: dictionary with  attributes used to create bridge.
    :param gateway: whether or not the bridge is a gateway.
    :param filtering: whether or not to create filters on the bridge.
    :param mtu: MTU of bridge.

    If net_attrs is set, it will add the net_attrs['gateway'] to the bridge
    using net_attrs['broadcast'] and net_attrs['cidr'].  It will also add
    the ip_v6 address specified in net_attrs['cidr_v6'] if use_ipv6 is set.

    The code will attempt to move any ips that already exist on the
    interface onto the bridge and reset the default gateway if necessary.

    """
    if not ip_lib.exists(bridge):
        LOG.debug('Starting Bridge %s', bridge)
        ip_lib.add(bridge, 'bridge')
        _disable_ipv6(bridge)
        _arp_filtering(bridge)
        ip_lib.set(bridge, state='up')

    if interface and ip_lib.exists(interface):
        LOG.debug('Adding interface %(interface)s to bridge %(bridge)s',
                  {'interface': interface, 'bridge': bridge})
        ip_lib.set(interface, master=bridge, state='up',
                   check_exit_code=[0, 2, 254])

        _set_device_mtu(interface, mtu)
        _update_bridge_routes(interface, bridge)
        # NOTE(sean-k-mooney):
        # The bridge mtu cannot be set until after an
        # interface is added due to bug:
        # https://bugs.launchpad.net/ubuntu/+source/linux/+bug/1399064
        _set_device_mtu(bridge, mtu)


def _ensure_bridge_filtering(bridge, gateway):
    # This method leaves privsep usage to iptables manager
    # Don't forward traffic unless we were told to be a gateway
    LOG.debug("Ensuring filtering %s to %s", bridge, gateway)
    global _IPTABLES_MANAGER
    ipv4_filter = _IPTABLES_MANAGER.ipv4['filter']
    if gateway:
        for rule in _IPTABLES_MANAGER.get_gateway_rules(bridge):
            ipv4_filter.add_rule(*rule)
    else:
        ipv4_filter.add_rule('FORWARD',
                             ('--in-interface %s -j %s'
                              % (bridge,
                                 _IPTABLES_MANAGER.iptables_drop_action)))
    ipv4_filter.add_rule('FORWARD',
                         ('--out-interface %s -j %s'
                          % (bridge,
                             _IPTABLES_MANAGER.iptables_drop_action)))
    _IPTABLES_MANAGER.apply()


def configure(iptables_mgr):
    """Configure the iptables manager impl.

    :param iptables_mgr: the iptables manager instance
    """
    global _IPTABLES_MANAGER
    _IPTABLES_MANAGER = iptables_mgr



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

import testtools

from os_vif import objects

from vif_plug_linux_bridge import constants
from vif_plug_linux_bridge import linux_bridge
from vif_plug_linux_bridge import linux_net


class PluginTest(testtools.TestCase):

    def __init__(self, *args, **kwargs):
        super(PluginTest, self).__init__(*args, **kwargs)

        objects.register_all()

        self.instance = objects.instance_info.InstanceInfo(
            name='demo',
            uuid='f0000000-0000-0000-0000-000000000001')

    @mock.patch.object(linux_net, 'ensure_vlan_bridge')
    @mock.patch.object(linux_net, 'ensure_bridge')
    def test_plug_bridge(self, mock_ensure_bridge,
                         mock_ensure_vlan_bridge):
        network = objects.network.Network(
            id='437c6db5-4e6f-4b43-b64b-ed6a11ee5ba7',
            bridge='br0')

        vif = objects.vif.VIFBridge(
            id='b679325f-ca89-4ee0-a8be-6db1409b69ea',
            address='ca:fe:de:ad:be:ef',
            network=network,
            dev_name='tap-xxx-yyy-zzz',
            bridge_name="br0")

        plugin = linux_bridge.LinuxBridgePlugin.load(constants.PLUGIN_NAME)
        plugin.plug(vif, self.instance)

        mock_ensure_bridge.assert_not_called()
        mock_ensure_vlan_bridge.assert_not_called()

    def test_plug_bridge_create_br_mtu_in_model(self):
        self._test_plug_bridge_create_br(mtu=1234)

    def test_plug_bridge_create_br_mtu_from_config(self):
        self._test_plug_bridge_create_br()

    @mock.patch.object(linux_net, 'ensure_vlan_bridge')
    @mock.patch.object(linux_net, 'ensure_bridge')
    def _test_plug_bridge_create_br(self, mock_ensure_bridge,
                                   mock_ensure_vlan_bridge,
                                   mtu=None):
        network = objects.network.Network(
            id='437c6db5-4e6f-4b43-b64b-ed6a11ee5ba7',
            bridge='br0',
            bridge_interface='eth0',
            should_provide_bridge=True,
            mtu=mtu)

        vif = objects.vif.VIFBridge(
            id='b679325f-ca89-4ee0-a8be-6db1409b69ea',
            address='ca:fe:de:ad:be:ef',
            network=network,
            dev_name='tap-xxx-yyy-zzz',
            has_traffic_filtering=True,
            bridge_name="br0")

        plugin = linux_bridge.LinuxBridgePlugin.load(constants.PLUGIN_NAME)
        plugin.plug(vif, self.instance)

        mock_ensure_bridge.assert_called_with("br0", "eth0",
                                              filtering=False,
                                              mtu=mtu or 1500)
        mock_ensure_vlan_bridge.assert_not_called()

        mock_ensure_bridge.reset_mock()
        vif.has_traffic_filtering = False
        plugin.plug(vif, self.instance)
        mock_ensure_bridge.assert_called_with("br0", "eth0",
                                              filtering=True,
                                              mtu=mtu or 1500)

    def test_plug_bridge_create_br_vlan_mtu_in_model(self):
        self._test_plug_bridge_create_br_vlan(mtu=1234)

    def test_plug_bridge_create_br_vlan_mtu_from_config(self):
        self._test_plug_bridge_create_br_vlan()

    @mock.patch.object(linux_net, 'ensure_vlan_bridge')
    @mock.patch.object(linux_net, 'ensure_bridge')
    def _test_plug_bridge_create_br_vlan(self, mock_ensure_bridge,
                                         mock_ensure_vlan_bridge,
                                         mtu=None):
        network = objects.network.Network(
            id='437c6db5-4e6f-4b43-b64b-ed6a11ee5ba7',
            bridge='br0',
            bridge_interface='eth0',
            vlan=99,
            should_provide_bridge=True,
            should_provide_vlan=True,
            mtu=mtu)

        vif = objects.vif.VIFBridge(
            id='b679325f-ca89-4ee0-a8be-6db1409b69ea',
            address='ca:fe:de:ad:be:ef',
            network=network,
            dev_name='tap-xxx-yyy-zzz',
            bridge_name="br0")

        plugin = linux_bridge.LinuxBridgePlugin.load(constants.PLUGIN_NAME)
        plugin.plug(vif, self.instance)

        mock_ensure_bridge.assert_not_called()
        mock_ensure_vlan_bridge.assert_called_with(
            99, "br0", "eth0", mtu=mtu or 1500)



# Copyright (c) 2021 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

__all__ = [
    'list_plugins_opts',
]

from copy import deepcopy
import os_vif


os_vif.initialize()

_EXT_MANAGER = os_vif._EXT_MANAGER

plugins_list = [
    (name, _EXT_MANAGER[name].obj)
    for name in sorted(_EXT_MANAGER.names())
]


def list_plugins_opts():
    return [('os_vif_' + g, deepcopy(o.CONFIG_OPTS)) for g, o in plugins_list]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from oslo_utils import importutils


interface_map = {
    'vsctl': 'vif_plug_ovs.ovsdb.impl_vsctl',
    'native': 'vif_plug_ovs.ovsdb.impl_idl',
}


def get_instance(context, iface_name=None):
    """Return the configured OVSDB API implementation"""
    iface = importutils.import_module(
        interface_map[iface_name or context.interface])
    return iface.api_factory(context)


class ImplAPI(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def has_table_column(self, table, column):
        """Check if a column exists in a database table

        :param table: (string) table name
        :param column: (string) column name
        :return: True if the column exists, False if not.
        """



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from os_vif import exception
from os_vif.tests.unit import base

"""Mostly inspired by os-brick's tests."""


class VIFExceptionTestCase(base.TestCase):
    def test_default_error_msg(self):
        class FakeVIFException(exception.ExceptionBase):
            msg_fmt = "default message"

        exc = FakeVIFException()
        self.assertEqual(str(exc), 'default message')

    def test_error_msg(self):
        self.assertEqual(str(exception.ExceptionBase('test')),
                         'test')

    def test_default_error_msg_with_kwargs(self):
        class FakeVIFException(exception.ExceptionBase):
            msg_fmt = "default message: %(foo)s"

        exc = FakeVIFException(foo="bar")
        self.assertEqual(str(exc), 'default message: bar')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import re
import time

from oslo_concurrency import processutils
from oslo_utils import excutils

from os_vif.internal.ip.api import ip as ip_lib
from os_vif.tests.functional import base
from os_vif.tests.functional import privsep


@privsep.os_vif_pctxt.entrypoint
def _execute_command(*args):
    return processutils.execute(*args)


class ShellIpCommands(object):

    def add_device(self, device, dev_type, peer=None, link=None,
                   vlan_id=None):
        if 'vlan' == dev_type:
            _execute_command('ip', 'link', 'add', 'link', link,
                             'name', device, 'type', dev_type, 'vlan', 'id',
                             vlan_id)
        elif 'veth' == dev_type:
            _execute_command('ip', 'link', 'add', device, 'type', dev_type,
                             'peer', 'name', peer)
        elif 'dummy' == dev_type:
            _execute_command('ip', 'link', 'add', device, 'type', dev_type)
        # ensure that the device exists to prevent racing
        # with other ip commands
        for _ in range(10):
            if self.exist_device(device):
                return
            time.sleep(0.1)

    def del_device(self, device):
        if self.exist_device(device):
            _execute_command('ip', 'link', 'del', device)

    def set_status_up(self, device):
        _execute_command('ip', 'link', 'set', device, 'up')

    def set_status_down(self, device):
        _execute_command('ip', 'link', 'set', device, 'down')

    def set_device_mtu(self, device, mtu):
        _execute_command('ip', 'link', 'set', device, 'mtu', mtu)

    def show_device(self, device):
        val, err = _execute_command('ip', 'link', 'show', device)
        return val.splitlines()

    def exist_device(self, device):
        try:
            _execute_command('ip', 'link', 'show', device)
            return True
        except processutils.ProcessExecutionError as e:
            with excutils.save_and_reraise_exception() as saved_exception:
                if e.exit_code == 1:
                    saved_exception.reraise = False
                    return False

    def show_state(self, device):
        regex = re.compile(r".*state (?P<state>\w+)")
        match = regex.match(self.show_device(device)[0])
        if match is None:
            return
        return match.group('state')

    def show_promisc(self, device):
        regex = re.compile(r".*(PROMISC)")
        match = regex.match(self.show_device(device)[0])
        return True if match else False

    def show_mac(self, device):
        exp = r".*link/ether (?P<mac>([0-9A-Fa-f]{2}[:]){5}[0-9A-Fa-f]{2})"
        regex = re.compile(exp)
        match = regex.match(self.show_device(device)[1])
        if match is None:
            return
        return match.group('mac')

    def show_mtu(self, device):
        regex = re.compile(r".*mtu (?P<mtu>\d+)")
        match = regex.match(self.show_device(device)[0])
        if match is None:
            return
        return int(match.group('mtu'))


@privsep.os_vif_pctxt.entrypoint
def _ip_cmd_set(*args, **kwargs):
    ip_lib.set(*args, **kwargs)


@privsep.os_vif_pctxt.entrypoint
def _ip_cmd_add(*args, **kwargs):
    ip_lib.add(*args, **kwargs)


@privsep.os_vif_pctxt.entrypoint
def _ip_cmd_delete(*args, **kwargs):
    ip_lib.delete(*args, **kwargs)


@privsep.os_vif_pctxt.entrypoint
def _ip_cmd_exists(*args, **kwargs):
    return ip_lib.exists(*args, **kwargs)


class TestIpCommand(ShellIpCommands, base.BaseFunctionalTestCase):

    def setUp(self):
        super(TestIpCommand, self).setUp()

    def test_set_state(self):
        device1 = "test_dev_1"
        device2 = "test_dev_2"
        self.addCleanup(self.del_device, device1)
        self.add_device(device1, 'veth', peer=device2)
        _ip_cmd_set(device1, state='up')
        _ip_cmd_set(device2, state='up')
        self.assertEqual('UP', self.show_state(device1))
        self.assertEqual('UP', self.show_state(device2))
        _ip_cmd_set(device1, state='down')
        _ip_cmd_set(device2, state='down')
        self.assertEqual('DOWN', self.show_state(device1))
        self.assertEqual('DOWN', self.show_state(device2))

    def test_set_mtu(self):
        device = "test_dev_3"
        self.addCleanup(self.del_device, device)
        self.add_device(device, 'dummy')
        _ip_cmd_set(device, mtu=1200)
        self.assertEqual(1200, self.show_mtu(device))
        _ip_cmd_set(device, mtu=900)
        self.assertEqual(900, self.show_mtu(device))

    def test_set_address(self):
        device = "test_dev_4"
        address1 = "36:a7:e4:f9:01:01"
        address2 = "36:a7:e4:f9:01:01"
        self.addCleanup(self.del_device, device)
        self.add_device(device, 'dummy')
        _ip_cmd_set(device, address=address1)
        self.assertEqual(address1, self.show_mac(device))
        _ip_cmd_set(device, address=address2)
        self.assertEqual(address2, self.show_mac(device))

    def test_set_promisc(self):
        device = "test_dev_5"
        self.addCleanup(self.del_device, device)
        self.add_device(device, 'dummy')
        _ip_cmd_set(device, promisc=True)
        self.assertTrue(self.show_promisc(device))
        _ip_cmd_set(device, promisc=False)
        self.assertFalse(self.show_promisc(device))

    def test_add_vlan(self):
        device = "test_dev_6"
        link = "test_devlink"
        self.addCleanup(self.del_device, device)
        self.addCleanup(self.del_device, link)
        self.add_device(link, 'dummy')
        _ip_cmd_add(device, 'vlan', link=link, vlan_id=100)
        self.assertTrue(self.exist_device(device))

    def test_add_veth(self):
        device = "test_dev_7"
        peer = "test_devpeer"
        self.addCleanup(self.del_device, device)
        _ip_cmd_add(device, 'veth', peer=peer)
        self.assertTrue(self.exist_device(device))
        self.assertTrue(self.exist_device(peer))

    def test_delete(self):
        device = "test_dev_8"
        self.addCleanup(self.del_device, device)
        self.add_device(device, 'dummy')
        self.assertTrue(self.exist_device(device))
        _ip_cmd_delete(device)
        self.assertFalse(self.exist_device(device))

    def test_iproute_object_closes_correctly(self):
        # NOTE(ralonsoh): check https://bugs.launchpad.net/os-vif/+bug/1807949
        device = "test_dev_9"
        link = "test_devlink_2"
        self.add_device(link, 'dummy')
        self.addCleanup(self.del_device, device)
        self.addCleanup(self.del_device, link)
        for _ in range(300):
            _ip_cmd_add(device, 'vlan', link=link, vlan_id=100)
            _ip_cmd_delete(device)

    def test_exists(self):
        device = "test_dev_10"
        self.addCleanup(self.del_device, device)
        self.add_device(device, 'dummy')
        self.assertTrue(_ip_cmd_exists(device))
        self.del_device(device)
        self.assertFalse(_ip_cmd_exists(device))

    def test_add_bridge(self):
        device = "test_dev_11"
        self.addCleanup(self.del_device, device)
        _ip_cmd_add(device, 'bridge')
        self.assertTrue(self.exist_device(device))
        base_path = "/sys/class/net/test_dev_11/bridge/%s"
        with open(base_path % "forward_delay", "r") as f:
            self.assertEqual("0", f.readline().rstrip('\n'))
        with open(base_path % "stp_state", "r") as f:
            self.assertEqual("0", f.readline().rstrip('\n'))
        with open(base_path % "multicast_snooping", "r") as f:
            self.assertEqual("0", f.readline().rstrip('\n'))
        with open(base_path % "ageing_time", "r") as f:
            value = int(f.readline().rstrip('\n'))
            # NOTE(sean-k-mooney): IEEE 8021-Q recommends that the default
            # ageing should be 300 and the linux kernel defaults to 300
            # via an unconditional define. As such we expect this to be
            # 300 however since services like network-manager could change
            # the default on bridge creation we check that if it is not 300
            # then the value should not be 0.
            self.assertTrue(300 == value or value != 0)

    def test_add_bridge_with_mac_ageing_0(self):
        device = "test_dev_12"
        self.addCleanup(self.del_device, device)
        _ip_cmd_add(device, 'bridge', ageing=0)
        self.assertTrue(self.exist_device(device))
        base_path = "/sys/class/net/test_dev_12/bridge/%s"
        with open(base_path % "forward_delay", "r") as f:
            self.assertEqual("0", f.readline().rstrip('\n'))
        with open(base_path % "stp_state", "r") as f:
            self.assertEqual("0", f.readline().rstrip('\n'))
        with open(base_path % "ageing_time", "r") as f:
            self.assertEqual("0", f.readline().rstrip('\n'))
        with open(base_path % "multicast_snooping", "r") as f:
            self.assertEqual("0", f.readline().rstrip('\n'))

    def test_add_port_to_bridge(self):
        device = "test_dev_13"
        bridge = "test_dev_14"
        self.addCleanup(self.del_device, device)
        self.addCleanup(self.del_device, bridge)
        self.add_device(device, 'dummy')
        _ip_cmd_add(bridge, 'bridge')
        self.assertTrue(self.exist_device(device))
        self.assertTrue(self.exist_device(bridge))
        _ip_cmd_set(device, master=bridge)
        path = "/sys/class/net/{}/brif/{}".format(bridge, device)
        self.assertTrue(os.path.exists(path))



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools
import socket

from ovs.db import idl
from ovs import socket_util
from ovs import stream
from ovsdbapp.backend.ovs_idl import connection
from ovsdbapp.backend.ovs_idl import idlutils
from ovsdbapp.backend.ovs_idl import vlog
from ovsdbapp.schema.open_vswitch import impl_idl

from vif_plug_ovs.ovsdb import api

REQUIRED_TABLES = ('Interface', 'Port', 'Bridge', 'Open_vSwitch', 'QoS')


def idl_factory(config):
    conn = config.connection
    schema_name = 'Open_vSwitch'
    helper = idlutils.get_schema_helper(conn, schema_name)
    for table in REQUIRED_TABLES:
        helper.register_table(table)
    return idl.Idl(conn, helper)


def api_factory(config):
    conn = connection.Connection(
        idl=idl_factory(config),
        timeout=config.timeout)
    return NeutronOvsdbIdl(conn)


class NeutronOvsdbIdl(impl_idl.OvsdbIdl, api.ImplAPI):
    """IDL interface for OVS database back-end

    This class provides an OVSDB IDL (Open vSwitch Database Interface
    Definition Language) interface to the OVS back-end.
    """

    def __init__(self, conn):
        vlog.use_python_logger()
        super(NeutronOvsdbIdl, self).__init__(conn)

    def _get_table_columns(self, table):
        return list(self.tables[table].columns)

    def has_table_column(self, table, column):
        return column in self._get_table_columns(table)


# this is derived form https://review.opendev.org/c/openstack/neutron/+/794892
def add_keepalives(fn):
    @functools.wraps(fn)
    def _open(*args, **kwargs):
        error, sock = fn(*args, **kwargs)
        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1)
        except socket.error as e:
            sock.close()
            return socket_util.get_exception_errno(e), None
        return error, sock
    return _open


class NoProbesMixin:
    @staticmethod
    def needs_probes():
        # If we are using keepalives, we can force probe_interval=0
        return False


class TCPStream(stream.TCPStream, NoProbesMixin):
    @classmethod
    @add_keepalives
    def _open(cls, suffix, dscp):
        return super()._open(suffix, dscp)


class SSLStream(stream.SSLStream, NoProbesMixin):
    @classmethod
    @add_keepalives
    def _open(cls, suffix, dscp):
        return super()._open(suffix, dscp)


# Overwriting globals in a library is clearly a good idea
stream.Stream.register_method("tcp", TCPStream)
stream.Stream.register_method("ssl", SSLStream)



#
# Copyright (C) 2016 Red Hat, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_privsep import capabilities as c
from oslo_privsep import priv_context

vif_plug = priv_context.PrivContext(
    "vif_plug_ovs",
    cfg_section="vif_plug_ovs_privileged",
    pypath=__name__ + ".vif_plug",
    capabilities=[
        c.CAP_NET_ADMIN,
    ],
)

vif_plug_test = priv_context.PrivContext(
    "vif_plug_ovs",
    cfg_section="vif_plug_ovs_privileged",
    pypath=__name__ + ".vif_plug_test",
    capabilities=[
        c.CAP_NET_ADMIN, c.CAP_DAC_OVERRIDE,
    ],
)



# Derived from nova/virt/libvirt/vif.py
#
# Copyright (C) 2011 Midokura KK
# Copyright (C) 2011 Nicira, Inc
# Copyright 2011 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import debtcollector.removals
from os_vif import objects
from os_vif import plugin
from oslo_config import cfg

from vif_plug_linux_bridge import constants
from vif_plug_linux_bridge import iptables
from vif_plug_linux_bridge import linux_net


class LinuxBridgePlugin(plugin.PluginBase):
    """A VIF type that uses a standard Linux bridge device."""

    CONFIG_OPTS = (
        cfg.BoolOpt('use_ipv6',
                    default=False,
                    help='Use IPv6',
                    deprecated_group="DEFAULT"),

        cfg.StrOpt('iptables_top_regex',
                   default='',
                   help='Regular expression to match the iptables rule that '
                   'should always be on the top.',
                   deprecated_group="DEFAULT"),
        cfg.StrOpt('iptables_bottom_regex',
                   default='',
                   help='Regular expression to match the iptables rule that '
                   'should always be on the bottom.',
                   deprecated_group="DEFAULT"),
        cfg.StrOpt('iptables_drop_action',
                   default='DROP',
                   help='The table that iptables to jump to when a packet is '
                   'to be dropped.',
                   deprecated_group="DEFAULT"),

        cfg.MultiStrOpt('forward_bridge_interface',
                        default=['all'],
                        help='An interface that bridges can forward to. If '
                        'this is set to all then all traffic will be '
                        'forwarded. Can be specified multiple times.',
                        deprecated_group="DEFAULT"),
        cfg.StrOpt('vlan_interface',
                   help='VLANs will bridge into this interface if set',
                   deprecated_group="DEFAULT"),
        cfg.StrOpt('flat_interface',
                   help='FlatDhcp will bridge into this interface if set',
                   deprecated_group="DEFAULT"),
        cfg.IntOpt('network_device_mtu',
                   default=1500,
                   help='MTU setting for network interface.',
                   deprecated_group="DEFAULT"),
    )

    def __init__(self, config):
        super(LinuxBridgePlugin, self).__init__(config)

        ipm = iptables.IptablesManager(
            use_ipv6=config.use_ipv6,
            iptables_top_regex=config.iptables_top_regex,
            iptables_bottom_regex=config.iptables_bottom_regex,
            iptables_drop_action=config.iptables_drop_action,
            forward_bridge_interface=config.forward_bridge_interface)

        linux_net.configure(ipm)

    def describe(self):
        return objects.host_info.HostPluginInfo(
            plugin_name=constants.PLUGIN_NAME,
            vif_info=[
                objects.host_info.HostVIFInfo(
                    vif_object_name=objects.vif.VIFBridge.__name__,
                    min_version="1.0",
                    max_version="1.0",
                    # NOTE(ralonsoh): currently 'supported_port_profiles' is
                    # only being used with OVS HostVIFInfo objects.
                    supported_port_profiles=[]),
            ])

    @debtcollector.removals.remove(
        message="LinuxBridgePlugin is deprecated and will be removed in "
                "a future release",
        category=DeprecationWarning)
    def plug(self, vif, instance_info):
        """Ensure that the bridge exists, and add VIF to it."""
        network = vif.network
        bridge_name = vif.bridge_name
        if not network.multi_host and network.should_provide_bridge:
            mtu = network.mtu or self.config.network_device_mtu
            if network.should_provide_vlan:
                iface = self.config.vlan_interface or network.bridge_interface
                linux_net.ensure_vlan_bridge(network.vlan,
                                             bridge_name, iface, mtu=mtu)
            else:
                iface = self.config.flat_interface or network.bridge_interface
                # only put in iptables rules if Neutron not filtering
                install_filters = not vif.has_traffic_filtering
                linux_net.ensure_bridge(bridge_name, iface,
                                        filtering=install_filters, mtu=mtu)

    @debtcollector.removals.remove(
        message="LinuxBridgePlugin is deprecated and will be removed in "
                "a future release",
        category=DeprecationWarning)
    def unplug(self, vif, instance_info):
        # Nothing required to unplug a port for a VIF using standard
        # Linux bridge device...
        pass



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

from unittest import mock

from oslo_serialization import jsonutils
from oslo_versionedobjects import base
from oslo_versionedobjects import fields

from os_vif.objects import base as osv_base
from os_vif.tests.unit import base as test_base


class TestVersionedObjectPrintable(test_base.TestCase):
    @base.VersionedObjectRegistry.register_if(False)
    class OVOChild1(osv_base.VersionedObject,
                    osv_base.VersionedObjectPrintableMixin):
        fields = {
            "child1_field1": fields.ListOfIntegersField()
        }

    @base.VersionedObjectRegistry.register_if(False)
    class OVOParent(osv_base.VersionedObject,
                    osv_base.VersionedObjectPrintableMixin,
                    base.ComparableVersionedObject):
        fields = {
            "parent_field1": fields.ListOfObjectsField("OVOChild1"),
            "parent_field2": fields.StringField(),
        }

    def setUp(self):
        super(TestVersionedObjectPrintable, self).setUp()
        child1_1 = self.OVOChild1(child1_field1=[1, 2, 3])
        child1_2 = self.OVOChild1(child1_field1=[4, 5, 6])
        self.obj = self.OVOParent(
            parent_field1=[child1_1, child1_2],
            parent_field2="test string")

    def test_print_object(self):
        out = str(self.obj)
        self.assertIn("'child1_field1': [1, 2, 3]}", out)
        self.assertIn("'child1_field1': [4, 5, 6]}", out)
        cmp = str({'parent_field2': "test string"})
        cmp = cmp.replace('{', '').replace('}', '')
        self.assertIn(str(cmp), out)

    @mock.patch.object(base.VersionedObject, "obj_class_from_name",
                       side_effect=[OVOParent, OVOChild1, OVOChild1])
    def test_serialize_object(self, *mock):
        """Test jsonutils serialization is not affected by this new mixin."""
        obj_orig = copy.deepcopy(self.obj)
        obj_orig_primitive = obj_orig.obj_to_primitive()
        str_orig_primitive = jsonutils.dumps(obj_orig_primitive)
        obj_new_primitive = jsonutils.loads(str_orig_primitive)
        obj_new = self.OVOParent.obj_from_primitive(obj_new_primitive)
        self.assertEqual(obj_orig_primitive, obj_new_primitive)
        self.assertEqual(obj_orig, obj_new)

    def test_import_non_ovo_class(self):
        """Test VersionedObjectPrintable could be inherited by non-OVO classes.
        """
        class NonOVOClass(osv_base.VersionedObjectPrintableMixin):
            def __str__(self):
                return "NonOVOClass __str__ method"

        self.assertEqual("NonOVOClass __str__ method", str(NonOVOClass()))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from unittest import mock

from pyroute2 import iproute
from pyroute2.netlink import exceptions as ipexc
from pyroute2.netlink.rtnl import ifinfmsg

from os_vif import exception
from os_vif.internal.ip.linux import impl_pyroute2
from os_vif.tests.unit import base


class TestIpCommand(base.TestCase):

    ERROR_CODE = 40
    OTHER_ERROR_CODE = 50
    DEVICE = 'device'
    MTU = 1500
    MAC = 'ca:fe:ca:fe:ca:fe'
    UP = 'up'
    TYPE_VETH = 'veth'
    TYPE_VLAN = 'vlan'
    TYPE_BRIDGE = 'bridge'
    LINK = 'device2'
    VLAN_ID = 14

    def setUp(self):
        super(TestIpCommand, self).setUp()
        self.ip = impl_pyroute2.PyRoute2()
        self.ip_link_p = mock.patch.object(iproute.IPRoute, 'link')
        self.ip_link = self.ip_link_p.start()

    def test_set(self):
        with mock.patch.object(iproute.IPRoute, 'link_lookup',
                               return_value=[1]) as mock_link_lookup:
            self.ip_link.return_value = [{'flags': 0x4000}]
            self.ip.set(self.DEVICE, state=self.UP, mtu=self.MTU,
                        address=self.MAC, promisc=True)
            mock_link_lookup.assert_called_once_with(ifname=self.DEVICE)
            args = {'state': self.UP,
                    'mtu': self.MTU,
                    'address': self.MAC,
                    'flags': 0x4000 | ifinfmsg.IFF_PROMISC}
            calls = [mock.call('get', index=1),
                     mock.call('set', index=1, **args)]
            self.ip_link.assert_has_calls(calls)

    def test_set_exit_code(self):
        with mock.patch.object(iproute.IPRoute, 'link_lookup',
                               return_value=[1]) as mock_link_lookup:
            self.ip_link.side_effect = ipexc.NetlinkError(self.ERROR_CODE,
                                                          msg="Error message")

            self.ip.set(self.DEVICE, check_exit_code=[self.ERROR_CODE])
            mock_link_lookup.assert_called_once_with(ifname=self.DEVICE)
            self.ip_link.assert_called_once_with('set', index=1)

            self.assertRaises(ipexc.NetlinkError, self.ip.set, self.DEVICE,
                              check_exit_code=[self.OTHER_ERROR_CODE])

    def test_set_no_interface_found(self):
        with mock.patch.object(iproute.IPRoute, 'link_lookup',
                               return_value=[]) as mock_link_lookup:
            self.assertRaises(exception.NetworkInterfaceNotFound, self.ip.set,
                              self.DEVICE)
            mock_link_lookup.assert_called_once_with(ifname=self.DEVICE)
            self.ip_link.assert_not_called()

    def test_add_veth(self):
        self.ip.add(self.DEVICE, self.TYPE_VETH, peer='peer')
        self.ip_link.assert_called_once_with(
            'add', ifname=self.DEVICE, kind=self.TYPE_VETH, peer='peer')

    def test_add_vlan(self):
        with mock.patch.object(iproute.IPRoute, 'link_lookup',
                               return_value=[1]) as mock_link_lookup:
            self.ip.add(self.DEVICE, self.TYPE_VLAN, link=self.LINK,
                        vlan_id=self.VLAN_ID)
            mock_link_lookup.assert_called_once_with(ifname=self.LINK)
            args = {'ifname': self.DEVICE,
                    'kind': self.TYPE_VLAN,
                    'vlan_id': self.VLAN_ID,
                    'link': 1}
            self.ip_link.assert_called_once_with('add', **args)

    def test_add_bridge(self):
        self.ip.add(self.DEVICE, self.TYPE_BRIDGE)
        args = {'ifname': self.DEVICE,
                'kind': self.TYPE_BRIDGE,
                'IFLA_BR_FORWARD_DELAY': 0,
                'IFLA_BR_STP_STATE': 0,
                'IFLA_BR_MCAST_SNOOPING': 0}
        self.ip_link.assert_called_once_with('add', **args)

    def test_add_bridge_with_ageing(self):
        self.ip.add(self.DEVICE, self.TYPE_BRIDGE, ageing=0)
        args = {'ifname': self.DEVICE,
                'kind': self.TYPE_BRIDGE,
                'IFLA_BR_AGEING_TIME': 0,
                'IFLA_BR_FORWARD_DELAY': 0,
                'IFLA_BR_STP_STATE': 0,
                'IFLA_BR_MCAST_SNOOPING': 0}
        self.ip_link.assert_called_once_with('add', **args)

    def test_add_vlan_no_interface_found(self):
        with mock.patch.object(iproute.IPRoute, 'link_lookup',
                               return_value=[]) as mock_link_lookup:
            self.assertRaises(exception.NetworkInterfaceNotFound, self.ip.add,
                              self.DEVICE, self.TYPE_VLAN, link=self.LINK)
            mock_link_lookup.assert_called_once_with(ifname=self.LINK)
            self.ip_link.assert_not_called()

    def test_add_other_type(self):
        self.assertRaises(exception.NetworkInterfaceTypeNotDefined,
                          self.ip.add, self.DEVICE, 'type_not_defined')

    def test_add_exit_code(self):
        self.ip_link.side_effect = ipexc.NetlinkError(self.ERROR_CODE,
                                                      msg="Error message")

        self.ip.add(self.DEVICE, self.TYPE_VETH, peer='peer',
                    check_exit_code=[self.ERROR_CODE])
        self.ip_link.assert_called_once_with(
            'add', ifname=self.DEVICE, kind=self.TYPE_VETH, peer='peer')

        self.assertRaises(
            exception.NetworkInterfaceNotFound,
            self.ip.add, self.DEVICE,
            self.TYPE_VLAN, peer='peer',
            check_exit_code=[self.OTHER_ERROR_CODE])

    def test_delete(self):
        with mock.patch.object(iproute.IPRoute, 'link_lookup',
                               return_value=[1]) as mock_link_lookup:
            self.ip.delete(self.DEVICE)
            mock_link_lookup.assert_called_once_with(ifname=self.DEVICE)
            self.ip_link.assert_called_once_with('del', index=1)

    def test_delete_no_interface_found(self):
        with mock.patch.object(iproute.IPRoute, 'link_lookup',
                               return_value=[]) as mock_link_lookup:
            self.assertRaises(exception.NetworkInterfaceNotFound,
                              self.ip.delete, self.DEVICE)
            mock_link_lookup.assert_called_once_with(ifname=self.DEVICE)

    def test_delete_exit_code(self):
        with mock.patch.object(iproute.IPRoute, 'link_lookup',
                               return_value=[1]) as mock_link_lookup:
            self.ip_link.side_effect = ipexc.NetlinkError(self.ERROR_CODE,
                                                          msg="Error message")

            self.ip.delete(self.DEVICE, check_exit_code=[self.ERROR_CODE])
            mock_link_lookup.assert_called_once_with(ifname=self.DEVICE)
            self.ip_link.assert_called_once_with('del', index=1)

            self.assertRaises(ipexc.NetlinkError, self.ip.delete, self.DEVICE,
                              check_exit_code=[self.OTHER_ERROR_CODE])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import glob
import os.path

from unittest import mock

import testtools

from os_vif.internal.ip.api import ip as ip_lib

from vif_plug_ovs import exception
from vif_plug_ovs import linux_net
from vif_plug_ovs import privsep


class LinuxNetTest(testtools.TestCase):

    def setUp(self):
        super(LinuxNetTest, self).setUp()

        privsep.vif_plug.set_client_mode(False)

    @mock.patch.object(linux_net, "_arp_filtering")
    @mock.patch.object(linux_net, "set_interface_state")
    @mock.patch.object(linux_net, "_disable_ipv6")
    @mock.patch.object(ip_lib, "add")
    @mock.patch.object(ip_lib, "exists", return_value=False)
    def test_ensure_bridge(self, mock_dev_exists, mock_add,
                           mock_disable_ipv6, mock_set_state,
                           mock_arp_filtering):
        linux_net.ensure_bridge("br0")

        mock_dev_exists.assert_called_once_with("br0")
        mock_add.assert_called_once_with("br0", "bridge", ageing=0)
        mock_disable_ipv6.assert_called_once_with("br0")
        mock_set_state.assert_called_once_with("br0", "up")
        mock_arp_filtering.assert_called_once_with("br0")

    @mock.patch.object(linux_net, "_arp_filtering")
    @mock.patch.object(linux_net, "set_interface_state")
    @mock.patch.object(linux_net, "_disable_ipv6")
    @mock.patch.object(ip_lib, "add")
    @mock.patch.object(ip_lib, "exists", return_value=True)
    def test_ensure_bridge_exists(self, mock_dev_exists, mock_add,
                                  mock_disable_ipv6, mock_set_state,
                                  mock_arp_filtering):
        linux_net.ensure_bridge("br0")

        mock_dev_exists.assert_called_once_with("br0")
        mock_add.assert_not_called()
        mock_disable_ipv6.assert_called_once_with("br0")
        mock_set_state.assert_called_once_with("br0", "up")
        mock_arp_filtering.assert_called_once_with("br0")

    @mock.patch('builtins.open')
    @mock.patch("os.path.exists")
    def test__disable_ipv6(self, mock_exists, mock_open):

        exists_path = "/proc/sys/net/ipv6/conf/br0/disable_ipv6"
        mock_exists.return_value = False

        linux_net._disable_ipv6("br0")
        mock_exists.assert_called_once_with(exists_path)
        mock_open.assert_not_called()

        mock_exists.reset_mock()
        mock_exists.return_value = True
        linux_net._disable_ipv6("br0")
        mock_exists.assert_called_once_with(exists_path)
        mock_open.assert_called_once_with(exists_path, 'w')

    @mock.patch.object(os.path, 'exists', return_value=True)
    @mock.patch('builtins.open')
    def test__arp_filtering(self, mock_open, *args):
        mock_open.side_effect = mock.mock_open()
        linux_net._arp_filtering("br0")

        mock_open.assert_has_calls([
            mock.call('/proc/sys/net/ipv4/conf/br0/arp_ignore', 'w'),
            mock.call('/proc/sys/net/ipv4/conf/br0/arp_announce', 'w')])
        mock_open.side_effect.return_value.write.assert_has_calls([
            mock.call('1'),
            mock.call('2')])

    @mock.patch.object(ip_lib, "delete")
    @mock.patch.object(ip_lib, "exists", return_value=False)
    def test_delete_bridge_none(self, mock_dev_exists, mock_delete):
        linux_net.delete_bridge("br0", "vnet1")

        mock_delete.assert_not_called()
        mock_dev_exists.assert_called_once_with("br0")

    @mock.patch.object(linux_net, "set_interface_state")
    @mock.patch.object(ip_lib, "delete")
    @mock.patch.object(ip_lib, "exists", return_value=True)
    def test_delete_bridge_exists(self, mock_dev_exists, mock_delete,
                                  mock_set_state):
        linux_net.delete_bridge("br0", "vnet1")

        mock_dev_exists.assert_has_calls([mock.call("br0"),
                                          mock.call("vnet1")])
        mock_delete.assert_called_once_with("br0", check_exit_code=[0, 2, 254])
        mock_set_state.assert_called_once_with("vnet1", "down")

    @mock.patch.object(linux_net, "set_interface_state")
    @mock.patch.object(ip_lib, "delete")
    @mock.patch.object(ip_lib, "exists")
    def test_delete_interface_not_present(self, mock_dev_exists, mock_delete,
                                          mock_set_state):
        mock_dev_exists.return_value = next(lambda: (yield True),
                                            (yield False))

        linux_net.delete_bridge("br0", "vnet1")

        mock_dev_exists.assert_has_calls([mock.call("br0"),
                                          mock.call("vnet1")])
        mock_delete.assert_called_once_with("br0", check_exit_code=[0, 2, 254])
        mock_set_state.assert_not_called()

    @mock.patch.object(ip_lib, "set")
    def test_add_bridge_port(self, mock_set):
        linux_net.add_bridge_port("br0", "vnet1")
        mock_set.assert_called_once_with("vnet1", master="br0")

    @mock.patch.object(linux_net, '_get_phys_switch_id')
    def test_is_switchdev_ioerror(self, mock__get_phys_switch_id):
        mock__get_phys_switch_id.side_effect = ([IOError()])
        test_switchdev = linux_net._is_switchdev('pf_ifname')
        self.assertEqual(test_switchdev, False)

    @mock.patch.object(linux_net, '_get_phys_switch_id')
    def test_is_switchdev_empty(self, mock__get_phys_switch_id):
        mock__get_phys_switch_id.return_value = ''
        test_switchdev = linux_net._is_switchdev('pf_ifname')
        self.assertEqual(test_switchdev, False)

    @mock.patch.object(linux_net, '_get_phys_switch_id')
    def test_is_switchdev_positive(self, mock__get_phys_switch_id):
        mock__get_phys_switch_id.return_value = 'pf_sw_id'
        test_switchdev = linux_net._is_switchdev('pf_ifname')
        self.assertEqual(test_switchdev, True)

    def test_parse_vf_number(self):
        self.assertEqual(linux_net._parse_vf_number("0"), "0")
        self.assertEqual(linux_net._parse_vf_number("pf13vf42"), "42")
        self.assertEqual(linux_net._parse_vf_number("VF19@PF13"), "19")
        self.assertIsNone(linux_net._parse_vf_number("p7"))
        self.assertIsNone(linux_net._parse_vf_number("pf31"))
        self.assertIsNone(linux_net._parse_vf_number("g4rbl3d"))

    def test_parse_pf_number(self):
        self.assertIsNone(linux_net._parse_pf_number("0"))
        self.assertEqual(linux_net._parse_pf_number("pf13vf42"), "13")
        self.assertEqual(linux_net._parse_pf_number("VF19@PF13"), "13")
        self.assertIsNone(linux_net._parse_pf_number("p7"))
        self.assertEqual(linux_net._parse_pf_number("pf31"), "31")
        self.assertIsNone(linux_net._parse_pf_number("g4rbl3d"))

    @mock.patch.object(os, 'listdir')
    @mock.patch.object(linux_net, "_get_pf_func")
    @mock.patch.object(linux_net, "_get_phys_port_name")
    @mock.patch.object(linux_net, '_get_phys_switch_id')
    def test_get_representor_port(self, mock__get_phys_switch_id,
                                  mock__get_phys_port_name,
                                  mock__get_pf_func,
                                  mock_listdir):
        mock_listdir.return_value = [
            'pf_ifname', 'rep_vf_1', 'rep_vf_2'
        ]
        mock__get_phys_switch_id.return_value = 'pf_sw_id'
        mock__get_pf_func.return_value = "0"
        mock__get_phys_port_name.side_effect = (['1', "pf0vf1", "pf0vf2"])
        ifname = linux_net.get_representor_port('pf_ifname', '2')
        self.assertEqual('rep_vf_2', ifname)

    @mock.patch.object(os, 'listdir')
    @mock.patch.object(linux_net, "_get_pf_func")
    @mock.patch.object(linux_net, "_get_phys_port_name")
    @mock.patch.object(linux_net, "_get_phys_switch_id")
    def test_get_representor_port_2_pfs(
            self, mock__get_phys_switch_id, mock__get_phys_port_name,
            mock__get_pf_func, mock_listdir):
        mock_listdir.return_value = [
            'pf_ifname1', 'pf_ifname2', 'rep_pf1_vf_1', 'rep_pf1_vf_2',
            'rep_pf2_vf_1', 'rep_pf2_vf_2',
        ]
        mock__get_phys_switch_id.return_value = 'pf_sw_id'
        mock__get_pf_func.return_value = "2"
        mock__get_phys_port_name.side_effect = (
            ["p1", "p2", "VF1@PF1", "pf2vf1", "vf2@pf1", "pf2vf2"])
        ifname = linux_net.get_representor_port('pf_ifname2', '2')
        self.assertEqual('rep_pf2_vf_2', ifname)

    @mock.patch.object(os, 'listdir')
    @mock.patch.object(linux_net, "_get_pf_func")
    @mock.patch.object(linux_net, "_get_phys_switch_id")
    @mock.patch.object(linux_net, "_get_phys_port_name")
    def test_get_representor_port_not_found(
            self, mock__get_phys_port_name, mock__get_phys_switch_id,
            mock__get_pf_func, mock_listdir):
        mock_listdir.return_value = [
            'pf_ifname', 'rep_vf_1', 'rep_vf_2'
        ]
        mock__get_phys_switch_id.return_value = 'pf_sw_id'
        mock__get_pf_func.return_value = "0"
        mock__get_phys_port_name.side_effect = (
            ["p0", "1", "2"])
        self.assertRaises(
            exception.RepresentorNotFound,
            linux_net.get_representor_port,
            'pf_ifname', '3'),

    @mock.patch.object(os, 'listdir')
    @mock.patch.object(linux_net, "_get_pf_func")
    @mock.patch.object(linux_net, "_get_phys_port_name")
    @mock.patch.object(linux_net, "_get_phys_switch_id")
    def test_get_representor_port_exception_io_error(
            self, mock__get_phys_switch_id, mock__get_phys_port_name,
            mock__get_pf_func, mock_listdir):
        mock_listdir.return_value = [
            'pf_ifname', 'rep_vf_1', 'rep_vf_2'
        ]
        mock__get_phys_switch_id.side_effect = (
            ['pf_sw_id', 'pf_sw_id', IOError(), 'pf_sw_id', '2'])
        mock__get_pf_func.return_value = "0"
        mock__get_phys_port_name.side_effect = (
            ["p0", "pf0vf0", "pf0vf1"])
        self.assertRaises(
            exception.RepresentorNotFound,
            linux_net.get_representor_port,
            'pf_ifname', '3')

    @mock.patch.object(os, 'listdir')
    @mock.patch.object(linux_net, "_get_pf_func")
    @mock.patch.object(linux_net, "_get_phys_port_name")
    @mock.patch.object(linux_net, "_get_phys_switch_id")
    def test_get_representor_port_exception_value_error(
            self, mock__get_phys_switch_id, mock__get_phys_port_name,
            mock__get_pf_func, mock_listdir):
        mock_listdir.return_value = [
            'pf_ifname', 'rep_vf_1', 'rep_vf_2'
        ]
        mock__get_phys_switch_id.return_value = 'pf_sw_id'
        mock__get_phys_port_name.side_effect = (['p0', '1', 'a'])
        mock__get_pf_func.return_value = "0"
        self.assertRaises(
            exception.RepresentorNotFound,
            linux_net.get_representor_port,
            'pf_ifname', '3')

    @mock.patch.object(os, 'listdir')
    @mock.patch.object(linux_net, '_get_phys_switch_id')
    @mock.patch.object(linux_net, "_get_phys_port_name")
    def test_physical_function_interface_name(
            self, mock__get_phys_port_name, mock__get_phys_switch_id,
            mock_listdir):
        mock_listdir.return_value = ['foo', 'bar']
        mock__get_phys_switch_id.side_effect = (
            ['', 'valid_switch'])
        mock__get_phys_port_name.side_effect = (["p1"])
        ifname = linux_net.get_ifname_by_pci_address(
            '0000:00:00.1', pf_interface=True, switchdev=False)
        self.assertEqual(ifname, 'foo')

    @mock.patch.object(os, 'listdir')
    @mock.patch.object(linux_net, "_get_phys_switch_id")
    @mock.patch.object(linux_net, "_get_phys_port_name")
    def test_physical_function_interface_name_with_switchdev(
            self, mock__get_phys_port_name, mock__get_phys_switch_id,
            mock_listdir):
        mock_listdir.return_value = ['foo', 'bar']
        mock__get_phys_switch_id.side_effect = (
            ['', 'valid_switch'])
        mock__get_phys_port_name.side_effect = (["p1s0"])
        ifname = linux_net.get_ifname_by_pci_address(
            '0000:00:00.1', pf_interface=True, switchdev=True)
        self.assertEqual(ifname, 'bar')

    @mock.patch.object(os, 'listdir')
    @mock.patch.object(linux_net, "_get_phys_switch_id")
    @mock.patch.object(linux_net, "_get_phys_port_name")
    def test_physical_function_interface_name_with_representors(
            self, mock__get_phys_port_name, mock__get_phys_switch_id,
            mock_listdir):
        # Get the PF that matches the phys_port_name regex
        mock_listdir.return_value = ['enp2s0f0_0', 'enp2s0f0_1', 'enp2s0f0']
        mock__get_phys_switch_id.side_effect = (
            ['valid_switch', 'valid_switch', 'valid_switch'])
        mock__get_phys_port_name.side_effect = (["pf0vf0", "pf0vf1", "p0"])
        ifname = linux_net.get_ifname_by_pci_address(
            '0000:00:00.1', pf_interface=True, switchdev=True)
        self.assertEqual(ifname, 'enp2s0f0')

    @mock.patch.object(os, 'listdir')
    @mock.patch.object(linux_net, "_get_phys_switch_id")
    @mock.patch.object(linux_net, "_get_phys_port_name")
    def test_physical_function_interface_name_with_fallback_To_first_netdev(
            self, mock__get_phys_port_name, mock__get_phys_switch_id,
            mock_listdir):
        # Try with switchdev mode to get PF but fail because there is no match
        # for the phys_port_name then fallback to first interface found
        mock_listdir.return_value = ['enp2s0f0_0', 'enp2s0f0_1', 'enp2s0f0']
        mock__get_phys_switch_id.side_effect = (['valid_switch',
                                                 'valid_switch',
                                                 'valid_switch'])
        mock__get_phys_port_name.side_effect = (["pf0vf0", "pf0vf1", "pf0vf2"])
        ifname = linux_net.get_ifname_by_pci_address(
            '0000:00:00.1', pf_interface=True, switchdev=True)
        self.assertEqual(ifname, 'enp2s0f0_0')

    @mock.patch.object(os, 'listdir')
    def test_get_ifname_by_pci_address_exception(self, mock_listdir):
        mock_listdir.side_effect = OSError('No such file or directory')
        self.assertRaises(
            exception.PciDeviceNotFoundById,
            linux_net.get_ifname_by_pci_address,
            '0000:00:00.1'
        )

    @mock.patch.object(os, 'readlink')
    @mock.patch.object(glob, 'iglob')
    def test_vf_number_found(self, mock_iglob, mock_readlink):
        mock_iglob.return_value = [
            '/sys/bus/pci/devices/0000:00:00.1/physfn/virtfn3',
        ]
        mock_readlink.return_value = '../../0000:00:00.1'
        vf_num = linux_net.get_vf_num_by_pci_address('0000:00:00.1')
        self.assertEqual(vf_num, '3')

    @mock.patch.object(os, 'readlink')
    @mock.patch.object(glob, 'iglob')
    def test_vf_number_not_found(self, mock_iglob, mock_readlink):
        mock_iglob.return_value = [
            '/sys/bus/pci/devices/0000:00:00.1/physfn/virtfn3',
        ]
        mock_readlink.return_value = '../../0000:00:00.2'
        self.assertRaises(
            exception.PciDeviceNotFoundById,
            linux_net.get_vf_num_by_pci_address,
            '0000:00:00.1'
        )

    @mock.patch.object(os, 'readlink')
    @mock.patch.object(glob, 'iglob')
    def test_get_vf_num_by_pci_address_exception(
            self, mock_iglob, mock_readlink):
        mock_iglob.return_value = [
            '/sys/bus/pci/devices/0000:00:00.1/physfn/virtfn3',
        ]
        mock_readlink.side_effect = OSError('No such file or directory')
        self.assertRaises(
            exception.PciDeviceNotFoundById,
            linux_net.get_vf_num_by_pci_address,
            '0000:00:00.1'
        )

    @mock.patch('builtins.open')
    @mock.patch.object(os.path, 'isfile')
    def test__get_phys_port_name(self, mock_isfile, mock_open):
        mock_open.return_value.__enter__ = lambda s: s
        readline_mock = mock_open.return_value.readline
        readline_mock.return_value = 'pf0vf0'
        mock_isfile.return_value = True
        phys_port_name = linux_net._get_phys_port_name("vf_ifname")
        self.assertEqual(phys_port_name, 'pf0vf0')

    @mock.patch.object(os.path, 'isfile')
    def test__get_phys_port_name_not_found(self, mock_isfile):
        mock_isfile.return_value = False
        phys_port_name = linux_net._get_phys_port_name("vf_ifname")
        self.assertIsNone(phys_port_name)

    @mock.patch('builtins.open')
    @mock.patch.object(os.path, 'isfile')
    def test__get_phys_switch_id(self, mock_isfile, mock_open):
        mock_open.return_value.__enter__ = lambda s: s
        readline_mock = mock_open.return_value.readline
        readline_mock.return_value = '66e40000039b0398'
        mock_isfile.return_value = True
        phys_port_name = linux_net._get_phys_switch_id("ifname")
        self.assertEqual(phys_port_name, '66e40000039b0398')

    @mock.patch.object(os.path, 'isfile')
    def test__get_phys_switch_id_not_found(self, mock_isfile):
        mock_isfile.return_value = False
        phys_port_name = linux_net._get_phys_switch_id("ifname")
        self.assertIsNone(phys_port_name)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import base as ovo_base


class VersionedObject(ovo_base.VersionedObject):

    OBJ_PROJECT_NAMESPACE = 'os_vif'


class VersionedObjectPrintableMixin(object):
    """Mix-in to implement __str__ method for a versioned object

    If a versioned object needs to be printable in a easy-reading format,
    inherit from this class.
    """

    def __str__(self):
        if callable(getattr(self, 'obj_to_primitive', None)):
            return str(self.obj_to_primitive())
        return super(VersionedObjectPrintableMixin, self).__str__()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from os_vif.i18n import _


class ExceptionBase(Exception):
    """Base Exception

    To correctly use this class, inherit from it and define
    a 'msg_fmt' property. That msg_fmt will get printf'd
    with the keyword arguments provided to the constructor.

    """
    msg_fmt = _("An unknown exception occurred.")

    def __init__(self, message=None, **kwargs):
        self.kwargs = kwargs

        if not message:
            try:
                message = self.msg_fmt % kwargs
            except Exception:
                # at least get the core message out if something happened
                message = self.msg_fmt

        self.message = message
        super(ExceptionBase, self).__init__(message)

    def format_message(self):
        # NOTE(mrodden): use the first argument to the python Exception object
        # which should be our full NovaException message, (see __init__)
        return self.args[0]


class LibraryNotInitialized(ExceptionBase):
    msg_fmt = _("Before using the os_vif library, you need to call "
                "os_vif.initialize()")


class NoMatchingPlugin(ExceptionBase):
    msg_fmt = _("No VIF plugin was found with the name %(plugin_name)s")


class NoMatchingPortProfileClass(ExceptionBase):
    msg_fmt = _("No PortProfile class was found with the name %(name)s")


class NoSupportedPortProfileVersion(ExceptionBase):
    msg_fmt = _("PortProfile class %(name)s "
                "versions %(got_versions)s do not satisfy "
                "min=%(min_version)s max=%(max_version)s")


class NoMatchingVIFClass(ExceptionBase):
    msg_fmt = _("No VIF class was found with the name %(name)s")


class NoSupportedVIFVersion(ExceptionBase):
    msg_fmt = _("VIF class %(name)s versions %(got_versions)s "
                "do not satisfy min=%(min_version)s max=%(max_version)s")


class PlugException(ExceptionBase):
    msg_fmt = _("Failed to plug VIF %(vif)s. Got error: %(err)s")


class UnplugException(ExceptionBase):
    msg_fmt = _("Failed to unplug VIF %(vif)s. Got error: %(err)s")


class NetworkMissingPhysicalNetwork(ExceptionBase):
    msg_fmt = _("Physical network is missing for network %(network_uuid)s")


class NetworkInterfaceNotFound(ExceptionBase):
    msg_fmt = _("Network interface %(interface)s not found")


class NetworkInterfaceTypeNotDefined(ExceptionBase):
    msg_fmt = _("Network interface type %(type)s not defined")


class ExternalImport(ExceptionBase):
    msg_fmt = _("Use of this module outside of os_vif is not allowed. It must "
                "not be imported in os-vif plugins that are out of tree as it "
                "is not a public interface of os-vif.")


class NotImplementedForOS(ExceptionBase):
    msg_fmt = _("Function %(function)s for %(os)s operating system")



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import testscenarios
import time
from unittest import mock

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_utils import uuidutils

from os_vif import objects

from vif_plug_ovs import constants
from vif_plug_ovs import ovs
from vif_plug_ovs.ovsdb import ovsdb_lib
from vif_plug_ovs import privsep
from vif_plug_ovs.tests.functional import base


CONF = cfg.CONF


@privsep.vif_plug.entrypoint
def run_privileged(*full_args):
    return processutils.execute(*full_args)[0].rstrip()


# derived from test_impl_pyroute2

def exist_device(device):
    try:
        run_privileged('ip', 'link', 'show', device)
        return True
    except processutils.ProcessExecutionError as e:
        if e.exit_code == 1:
            return False
        raise


def add_device(device, dev_type, peer=None, link=None,
                   vlan_id=None):
    if 'vlan' == dev_type:
        run_privileged('ip', 'link', 'add', 'link', link,
                         'name', device, 'type', dev_type, 'vlan', 'id',
                         vlan_id)
    elif 'veth' == dev_type:
        run_privileged('ip', 'link', 'add', device, 'type', dev_type,
                         'peer', 'name', peer)
    elif 'dummy' == dev_type:
        run_privileged('ip', 'link', 'add', device, 'type', dev_type)
    # ensure that the device exists to prevent racing with other ip commands
    for _ in range(10):
        if exist_device(device):
            return
        time.sleep(0.1)


def del_device(device):
    if exist_device(device):
        run_privileged('ip', 'link', 'del', device)


class TestOVSPlugin(testscenarios.WithScenarios,
                    base.VifPlugOvsBaseFunctionalTestCase):

    scenarios = [
        ('native', {'interface': 'native'}),
        ('vsctl', {'interface': 'vsctl'})
    ]

    def setUp(self):
        super(TestOVSPlugin, self).setUp()
        run_privileged('ovs-vsctl', 'set-manager', 'ptcp:6640')
        self.plugin = ovs.OvsPlugin.load(constants.PLUGIN_NAME)
        self.flags(ovsdb_interface=self.interface, group='os_vif_ovs')

        self.ovs = ovsdb_lib.BaseOVS(CONF.os_vif_ovs)
        self._ovsdb = self.ovs.ovsdb

        self.profile_ovs = objects.vif.VIFPortProfileOpenVSwitch(
            interface_id='e65867e0-9340-4a7f-a256-09af6eb7a3aa',
            datapath_type='netdev')

        self.subnet_bridge_4 = objects.subnet.Subnet(
            cidr='101.168.1.0/24',
            dns=['8.8.8.8'],
            gateway='101.168.1.1',
            dhcp_server='191.168.1.1')

        self.subnet_bridge_6 = objects.subnet.Subnet(
            cidr='101:1db9::/64',
            gateway='101:1db9::1')

        self.subnets = objects.subnet.SubnetList(
            objects=[self.subnet_bridge_4,
                     self.subnet_bridge_6])

        self.network_ovs_trunk = objects.network.Network(
            id='437c6db5-4e6f-4b43-b64b-ed6a11ee5ba7',
            bridge='%s01' % constants.TRUNK_BR_PREFIX,
            subnets=self.subnets,
            vlan=99)

        self.vif_vhostuser_trunk = objects.vif.VIFVHostUser(
            id='b679325f-ca89-4ee0-a8be-6db1409b69ea',
            address='ca:fe:de:ad:be:ef',
            network=self.network_ovs_trunk,
            path='/var/run/openvswitch/vhub679325f-ca',
            mode='client',
            port_profile=self.profile_ovs)

        self.profile_ovs_system = objects.vif.VIFPortProfileOpenVSwitch(
            interface_id='e65867e0-9340-4a7f-a256-09af6eb7a3aa',
            datapath_type='system',
            create_port=True)

        self.network_ovs = objects.network.Network(
            id='437c6db5-4e6f-4b43-b64b-ed6a11ee5ba7',
            bridge='br-qos-' + self.interface,
            subnets=self.subnets,
            vlan=99)

        self.vif_ovs_port = objects.vif.VIFOpenVSwitch(
            id='b679325f-ca89-4ee0-a8be-6db1409b69ea',
            address='ca:fe:de:ad:be:ef',
            network=self.network_ovs,
            port_profile=self.profile_ovs_system,
            vif_name="qos-port-" + self.interface)

        self.instance = objects.instance_info.InstanceInfo(
            name='demo',
            uuid='f0000000-0000-0000-0000-000000000001')

    def test_plug_unplug_ovs_vhostuser_trunk(self):
        trunk_bridge = '%s01' % constants.TRUNK_BR_PREFIX
        self.plugin.plug(self.vif_vhostuser_trunk, self.instance)
        self.addCleanup(self._del_bridge, trunk_bridge)
        self.assertTrue(self._check_bridge(trunk_bridge))

        other_bridge = 'br-%s' % uuidutils.generate_uuid()
        self._add_bridge(other_bridge)
        self.addCleanup(self._del_bridge, other_bridge)
        self.plugin.unplug(self.vif_vhostuser_trunk, self.instance)
        self.assertTrue(self._check_bridge(other_bridge))
        self.assertFalse(self._check_bridge(trunk_bridge))

    def test_plug_unplug_ovs_port_with_qos(self):
        bridge = 'br-qos-' + self.interface
        vif_name = "qos-port-" + self.interface
        qos_type = CONF.os_vif_ovs.default_qos_type
        self.addCleanup(self._del_bridge, bridge)
        self.addCleanup(
            self.ovs.delete_ovs_vif_port, bridge, vif_name,
            delete_netdev=False, qos_type=qos_type
        )
        self.addCleanup(del_device, vif_name)
        add_device(vif_name, 'dummy')
        # pluging a vif will create the port and bridge
        # if either does not exist
        self.plugin.plug(self.vif_ovs_port, self.instance)
        self.assertTrue(self._check_bridge(bridge))
        self.assertTrue(self._check_port(vif_name, bridge))
        qos_uuid = self.ovs.get_qos(
            vif_name, qos_type
        )[0]['_uuid']
        self._check_parameter('Port', vif_name, 'qos', qos_uuid)
        self._check_parameter(
            'QoS', str(qos_uuid), 'type', qos_type
        )
        # unpluging a port will not delete the bridge.
        self.plugin.unplug(self.vif_ovs_port, self.instance)
        self.assertTrue(self._check_bridge(bridge))
        self.assertFalse(self._check_port(vif_name, bridge))
        self._check_parameter(
            'QoS', str(qos_uuid), 'type', None
        )

    def test_plug_br_int_isolate_vif_dead_vlan(self):
        with mock.patch.object(self.plugin.config, 'isolate_vif', True):
            network = objects.network.Network(
                id='5449523c-3a08-11ef-86d6-17149687aa4d',
                bridge='br-5449523c',
                subnets=self.subnets,
                vlan=99)
            vif = objects.vif.VIFOpenVSwitch(
                id='85cb9bc6-3a08-11ef-b2d4-9b7c38edd677',
                address='ca:fe:de:ad:be:ef',
                network=network,
                port_profile=self.profile_ovs_system,
                vif_name="port-85cb9bc6")
            self.plugin.plug(vif, self.instance)
            self.addCleanup(self._del_bridge, 'br-5449523c')
            self._check_parameter('Port', vif.vif_name, 'tag', 4095)

    def test_plug_trunk_bridge_ignores_isolate_vif(self):
        with mock.patch.object(self.plugin.config, 'isolate_vif', True):
            network = objects.network.Network(
                id='ef98b384-3a0f-11ef-9009-47345fca266f',
                bridge='tbr-ef98b384',
                subnets=self.subnets,
                vlan=99)
            vif = objects.vif.VIFOpenVSwitch(
                id='631f52bc-3a07-11ef-a006-1319ef9d6edd',
                address='ca:fe:de:ad:be:ef',
                network=network,
                port_profile=self.profile_ovs_system,
                vif_name='port-631f52bc')
            self.plugin.plug(vif, self.instance)
            self.addCleanup(self._del_bridge, 'tbr-ef98b384')
            self._check_parameter('Port', vif.vif_name, 'tag', [])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import fixtures
import random


from unittest import mock

import testscenarios

from oslo_concurrency import processutils
from oslo_config import cfg
from oslo_utils import uuidutils

from vif_plug_ovs import constants
from vif_plug_ovs import linux_net
from vif_plug_ovs import ovs
from vif_plug_ovs.ovsdb import ovsdb_lib
from vif_plug_ovs import privsep
from vif_plug_ovs.tests.functional import base


CONF = cfg.CONF


@privsep.vif_plug_test.entrypoint
def run_privileged(*full_args):
    return processutils.execute(*full_args)[0].rstrip()


class TestOVSDBLib(testscenarios.WithScenarios,
                   base.VifPlugOvsBaseFunctionalTestCase):

    scenarios = [
        ('native', {'interface': 'native'}),
        ('vsctl', {'interface': 'vsctl'})
    ]

    def setUp(self):
        super(TestOVSDBLib, self).setUp()
        run_privileged('ovs-vsctl', 'set-manager', 'ptcp:6640')

        # NOTE: (ralonsoh) load default configuration variables "CONFIG_OPTS"
        ovs.OvsPlugin.load('ovs')
        self.flags(ovsdb_interface=self.interface, group='os_vif_ovs')

        self.ovs = ovsdb_lib.BaseOVS(CONF.os_vif_ovs)
        self._ovsdb = self.ovs.ovsdb
        self.brname = ('br' + str(random.randint(1000, 9999)) + '-' +
                       self.interface)

        # Make sure exceptions pass through by calling do_post_commit directly
        post_commit = (
            'ovsdbapp.schema.open_vswitch.impl_idl.'
            'OvsVsctlTransaction.post_commit'
        )
        # "this" is the self parmater which is a reference to the
        # OvsVsctlTransaction instance on which do_post_commit is defiend.

        def direct_post_commit(this, transaction):
            this.do_post_commit(transaction)

        self.useFixture(fixtures.MonkeyPatch(post_commit, direct_post_commit))

    def _add_port(self, bridge, port, may_exist=True):
        with self._ovsdb.transaction() as txn:
            txn.add(self._ovsdb.add_port(bridge, port, may_exist=may_exist))
            txn.add(self._ovsdb.db_set('Interface', port,
                                       ('type', 'internal')))
        self.assertIn(port, self._list_ports_in_bridge(bridge))

    def _list_ports_in_bridge(self, bridge):
        return self._ovsdb.list_ports(bridge).execute()

    def test__set_mtu_request(self):
        port_name = 'port1-' + self.interface
        self._add_bridge(self.brname)
        self.addCleanup(self._del_bridge, self.brname)
        self._add_port(self.brname, port_name)
        if self.ovs._ovs_supports_mtu_requests():
            with self._ovsdb.transaction() as txn:
                self.ovs._set_mtu_request(txn, port_name, 1000)
            self._check_parameter('Interface', port_name, 'mtu', 1000)
            with self._ovsdb.transaction() as txn:
                self.ovs._set_mtu_request(txn, port_name, 1500)
            self._check_parameter('Interface', port_name, 'mtu', 1500)
        else:
            self.skipTest('Current version of Open vSwitch does not support '
                          '"mtu_request" parameter')

    def test_create_ovs_vif_port(self):
        port_name = 'port2-' + self.interface
        iface_id = 'iface_id'
        mac = 'ca:fe:ca:fe:ca:fe'
        instance_id = uuidutils.generate_uuid()
        interface_type = constants.OVS_VHOSTUSER_INTERFACE_TYPE
        vhost_server_path = '/fake/path'
        mtu = 1500
        self._add_bridge(self.brname)
        self.addCleanup(self._del_bridge, self.brname)

        self.ovs.create_ovs_vif_port(self.brname, port_name, iface_id, mac,
                                     instance_id, mtu=mtu,
                                     interface_type=interface_type,
                                     vhost_server_path=vhost_server_path,
                                     tag=2000)

        expected_external_ids = {'iface-status': 'active',
                                 'iface-id': iface_id,
                                 'attached-mac': mac,
                                 'vm-uuid': instance_id}
        self._check_parameter('Interface', port_name, 'external_ids',
                              expected_external_ids)
        self._check_parameter('Interface', port_name, 'type', interface_type)
        expected_vhost_server_path = {'vhost-server-path': vhost_server_path}
        self._check_parameter(
            'Interface', port_name, 'options', expected_vhost_server_path
        )
        self._check_parameter('Port', port_name, 'tag', 2000)
        self._check_parameter('Port', port_name, 'qos', [])

    @mock.patch.object(linux_net, 'delete_net_dev')
    def test_delete_ovs_vif_port(self, *mock):
        port_name = 'port3-' + self.interface
        self._add_bridge(self.brname)
        self.addCleanup(self._del_bridge, self.brname)

        self._add_port(self.brname, port_name)
        self.ovs.delete_ovs_vif_port(self.brname, port_name)
        self.assertNotIn(port_name, self._list_ports_in_bridge(self.brname))

    def test_ensure_ovs_bridge(self):
        bridge_name = 'bridge2-' + self.interface
        self.ovs.ensure_ovs_bridge(bridge_name, constants.OVS_DATAPATH_SYSTEM)
        self.assertTrue(self._check_bridge(bridge_name))
        self.addCleanup(self._del_bridge, bridge_name)

    def test_create_patch_port_pair(self):
        port_bridge = 'fake-pb'
        port_bridge_port = 'fake-pbp'
        int_bridge = 'pb-int'
        int_bridge_port = 'fake-ibp'
        iface_id = 'iface_id'
        mac = 'ca:fe:ca:fe:ca:fe'
        instance_id = uuidutils.generate_uuid()

        # deleting a bridge deletes all ports on bridges so we register the
        # bridge cleanup first so if we fail anywhere it runs.
        self.addCleanup(self._del_bridge, port_bridge)
        self.addCleanup(self._del_bridge, int_bridge)
        self.ovs.ensure_ovs_bridge(port_bridge, constants.OVS_DATAPATH_SYSTEM)
        self.ovs.ensure_ovs_bridge(int_bridge, constants.OVS_DATAPATH_SYSTEM)
        self.ovs.create_patch_port_pair(
            port_bridge, port_bridge_port, int_bridge, int_bridge_port,
            iface_id, mac, instance_id, tag=2000)
        self.assertTrue(self._check_bridge(port_bridge))
        self.assertTrue(self._check_bridge(int_bridge))

        expected_external_ids = {'iface-status': 'active',
                                 'iface-id': iface_id,
                                 'attached-mac': mac,
                                 'vm-uuid': instance_id}
        self._check_parameter(
            'Interface', int_bridge_port, 'external_ids',
            expected_external_ids)
        self._check_parameter('Interface', int_bridge_port, 'type', 'patch')
        port_opts = {'peer': port_bridge_port}
        self._check_parameter(
            'Interface', int_bridge_port, 'options', port_opts)
        self._check_parameter('Port', int_bridge_port, 'tag', 2000)
        port_opts = {'peer': int_bridge_port}
        self._check_parameter(
            'Interface', port_bridge_port, 'options', port_opts)

    def test_create_ovs_vif_port_with_default_qos(self):
        if self.interface == 'native':
            self.skipTest(
                'test_create_ovs_vif_port_with_default_qos is unstable '
                'when run with the native driver, see: '
                'https://bugs.launchpad.net/os-vif/+bug/2087982')
        port_name = 'qos-port-' + self.interface
        iface_id = 'iface_id'
        mac = 'ca:fe:ca:fe:ca:fe'
        instance_id = uuidutils.generate_uuid()
        mtu = 1500
        interface_type = 'internal'
        qos_type = CONF.os_vif_ovs.default_qos_type

        self.addCleanup(self._del_bridge, self.brname)
        self._add_bridge(self.brname)

        self.addCleanup(
            self.ovs.delete_ovs_vif_port, self.brname, port_name,
            delete_netdev=False, qos_type=qos_type
        )
        self.ovs.create_ovs_vif_port(
            self.brname, port_name, iface_id, mac,
            instance_id, mtu=mtu, interface_type=interface_type,
            tag=2000, qos_type=qos_type
        )

        # first we assert that the standard parameters are set correctly
        expected_external_ids = {'iface-status': 'active',
                                 'iface-id': iface_id,
                                 'attached-mac': mac,
                                 'vm-uuid': instance_id}
        self._check_parameter('Interface', port_name, 'external_ids',
                              expected_external_ids)
        self._check_parameter('Interface', port_name, 'type', interface_type)
        self._check_parameter('Port', port_name, 'tag', 2000)

        # now we check that the port has a qos policy attached
        qos_uuid = self.ovs.get_qos(
            port_name, qos_type
        )[0]['_uuid']
        self._check_parameter('Port', port_name, 'qos', qos_uuid)

        # finally we check that the qos policy has the correct parameters
        self._check_parameter(
            'QoS', str(qos_uuid), 'type', qos_type
        )

    def test_delete_qos_if_exists(self):
        port_name = 'del-qos-port-' + self.interface
        iface_id = 'iface_id'
        mac = 'ca:fe:ca:fe:ca:fe'
        instance_id = uuidutils.generate_uuid()
        interface_type = 'internal'
        qos_type = CONF.os_vif_ovs.default_qos_type

        # setup test by creating a bridge and port, and register
        # cleanup funcitons to avoid leaking them.
        self.addCleanup(self._del_bridge, self.brname)
        self._add_bridge(self.brname)
        self.addCleanup(
            self.ovs.delete_ovs_vif_port, self.brname, port_name,
            delete_netdev=False, qos_type=qos_type
        )
        self.ovs.create_ovs_vif_port(
            self.brname, port_name, iface_id, mac,
            instance_id, interface_type=interface_type,
            qos_type=qos_type
        )

        # now we check that the port has a qos policy attached
        qos_uuid = self.ovs.get_qos(
            port_name, CONF.os_vif_ovs.default_qos_type
        )[0]['_uuid']
        self._check_parameter('Port', port_name, 'qos', qos_uuid)

        # finally we check that the qos policy has the correct parameters
        self._check_parameter(
            'QoS', str(qos_uuid), 'type', qos_type
        )

        # we need to delete the port directly in the db to remove
        # any references to the qos policy
        self.ovs.ovsdb.del_port(
            port_name, bridge=self.brname, if_exists=True).execute()
        # then we can delete the qos policy
        self.ovs.delete_qos_if_exists(port_name, qos_type)
        self._check_parameter(
            'QoS', str(qos_uuid), 'type', None
        )
        # invoking the delete when the policy does not exist
        # should not result in an error
        self.ovs.delete_qos_if_exists(port_name, qos_type)
        self._check_parameter(
            'QoS', str(qos_uuid), 'type', None
        )

    def test_get_qos(self):
        port_name = 'get-qos-' + self.interface
        iface_id = 'iface_id'
        mac = 'ca:fe:ca:fe:ca:fe'
        instance_id = uuidutils.generate_uuid()
        interface_type = 'internal'
        qos_type = CONF.os_vif_ovs.default_qos_type
        # initally no qos policy should exist
        self.assertEqual(0, len(self.ovs.get_qos(port_name, qos_type)))

        # if we create a port with a qos policy get_qos should
        # return the policy
        self.addCleanup(self._del_bridge, self.brname)
        self._add_bridge(self.brname)
        self.addCleanup(
            self.ovs.delete_ovs_vif_port, self.brname, port_name,
            delete_netdev=False, qos_type=qos_type
        )
        self.ovs.create_ovs_vif_port(
            self.brname, port_name, iface_id, mac,
            instance_id, interface_type=interface_type,
            qos_type=qos_type
        )
        # result should be a list of lenght 1 containing the
        # qos policy created for the port we defied.
        result = self.ovs.get_qos(port_name, qos_type)
        self.assertEqual(1, len(result))
        self.assertIn('_uuid', result[0])
        self._check_parameter(
            'Port', port_name, 'qos', result[0]['_uuid']
        )
        # if we delete the port and its qos policy get_qos should
        # not return it.
        self.ovs.delete_ovs_vif_port(
            self.brname, port_name,
            delete_netdev=False, qos_type=qos_type
        )
        self.assertEqual(0, len(self.ovs.get_qos(port_name, qos_type)))

    def test_port_exists(self):
        port_name = 'port-exists-' + self.interface
        iface_id = 'iface_id'
        mac = 'ca:fe:ca:fe:ca:fe'
        instance_id = uuidutils.generate_uuid()
        interface_type = 'internal'

        self.assertFalse(self.ovs.port_exists(port_name, self.brname))

        self.addCleanup(self._del_bridge, self.brname)
        self._add_bridge(self.brname)
        self.addCleanup(
            self.ovs.delete_ovs_vif_port, self.brname, port_name,
            delete_netdev=False,
        )
        self.ovs.create_ovs_vif_port(
            self.brname, port_name, iface_id, mac,
            instance_id, interface_type=interface_type,
        )

        self.assertTrue(self.ovs.port_exists(port_name, self.brname))



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'sphinx.ext.todo',
    'sphinx.ext.autodoc',
    'openstackdocstheme',
    'reno.sphinxext',
]

# openstackdocstheme options
openstackdocs_repo_name = 'openstack/os-vif'
openstackdocs_bug_project = 'os-vif'
openstackdocs_bug_tag = ''

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
copyright = '2016, OpenStack Foundation'

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'openstackdocs'



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import versionutils
from oslo_versionedobjects import base
from oslo_versionedobjects import fields

from os_vif import exception
from os_vif.objects import base as osv_base


def _get_common_version(object_name, max_version, min_version, exc_notmatch,
                        exc_notsupported):
    """Returns the accepted version from the loaded OVO registry"""
    reg = base.VersionedObjectRegistry.obj_classes()

    if object_name not in reg:
        raise exc_notmatch(name=object_name)

    gotvers = []
    for regobj in reg[object_name]:
        gotvers.append(regobj.VERSION)
        got = versionutils.convert_version_to_tuple(regobj.VERSION)
        minwant = versionutils.convert_version_to_tuple(min_version)
        maxwant = versionutils.convert_version_to_tuple(max_version)

        if minwant <= got <= maxwant:
            return regobj.VERSION

    raise exc_notsupported(name=object_name,
                           got_versions=",".join(gotvers),
                           min_version=min_version,
                           max_version=max_version)


@base.VersionedObjectRegistry.register
class HostPortProfileInfo(osv_base.VersionedObject,
                          base.ComparableVersionedObject,
                          osv_base.VersionedObjectPrintableMixin):
    """
    Class describing a PortProfile class and its supported versions
    """
    # Version 1.0: Initial version
    VERSION = "1.0"

    fields = {
        # object name of the subclass of os_vif.objects.vif.VIFPortProfileBase
        "profile_object_name": fields.StringField(),

        # String representing the earliest version of @name
        # that the plugin understands
        "min_version": fields.StringField(),

        # String representing the latest version of @name
        # that the plugin understands
        "max_version": fields.StringField(),
    }

    def get_common_version(self):
        return _get_common_version(self.profile_object_name,
                                   self.max_version,
                                   self.min_version,
                                   exception.NoMatchingPortProfileClass,
                                   exception.NoSupportedPortProfileVersion)


@base.VersionedObjectRegistry.register
class HostVIFInfo(osv_base.VersionedObject, base.ComparableVersionedObject,
                  osv_base.VersionedObjectPrintableMixin):
    """
    Class describing a VIF class and its supported versions
    """
    # Version 1.0: Initial version
    # Version 1.1: Adds 'supported_port_profiles' field
    VERSION = "1.1"

    fields = {
        # object name of the subclass of os_vif.objects.vif.VIFBase
        "vif_object_name": fields.StringField(),

        # String representing the earliest version of @name
        # that the plugin understands
        "min_version": fields.StringField(),

        # String representing the latest version of @name
        # that the plugin understands
        "max_version": fields.StringField(),

        # list of supported PortProfile objects and versions.
        "supported_port_profiles": fields.ListOfObjectsField(
            "HostPortProfileInfo")
    }

    def obj_make_compatible(self, primitive, target_version):
        target_version = versionutils.convert_version_to_tuple(target_version)
        if target_version < (1, 1) and 'supported_port_profiles' in primitive:
            del primitive['supported_port_profiles']
        super(HostVIFInfo, self).obj_make_compatible(primitive, '1.0')

    def get_common_version(self):
        return _get_common_version(self.vif_object_name,
                                   self.max_version,
                                   self.min_version,
                                   exception.NoMatchingVIFClass,
                                   exception.NoSupportedVIFVersion)


@base.VersionedObjectRegistry.register
class HostPluginInfo(osv_base.VersionedObject, base.ComparableVersionedObject,
                     osv_base.VersionedObjectPrintableMixin):
    """
    Class describing a plugin and its supported VIF classes
    """
    # Version 1.0: Initial version
    VERSION = "1.0"

    fields = {
        # name of the plugin
        "plugin_name": fields.StringField(),

        # list of HostVIFInfo instances supported by the plugin
        "vif_info": fields.ListOfObjectsField("HostVIFInfo"),
    }

    def has_vif(self, name):
        for vif in self.vif_info:
            if vif.vif_object_name == name:
                return True
        return False

    def get_vif(self, name):
        for vif in self.vif_info:
            if vif.vif_object_name == name:
                return vif

        raise exception.NoMatchingVIFClass(vif_name=name)

    def filter_vif_types(self, permitted_vif_type_names):
        new_vif_info = []
        for vif in self.vif_info:
            if vif.vif_object_name in permitted_vif_type_names:
                new_vif_info.append(vif)
        self.vif_info = new_vif_info


@base.VersionedObjectRegistry.register
class HostInfo(osv_base.VersionedObject, base.ComparableVersionedObject,
               osv_base.VersionedObjectPrintableMixin):
    """
    Class describing a host host and its supported plugin classes
    """

    fields = {
        # list of HostPluginInfo instances supported by the host host
        "plugin_info": fields.ListOfObjectsField("HostPluginInfo"),
    }

    def has_plugin(self, name):
        for plugin in self.plugin_info:
            if name == plugin.plugin_name:
                return True
        return False

    def get_plugin(self, name):
        for plugin in self.plugin_info:
            if name == plugin.plugin_name:
                return plugin

        raise exception.NoMatchingPlugin(plugin_name=name)

    def filter_vif_types(self, permitted_vif_type_names):
        new_plugins = []
        for plugin in self.plugin_info:
            plugin.filter_vif_types(permitted_vif_type_names)
            if len(plugin.vif_info) == 0:
                continue
            new_plugins.append(plugin)
        self.plugin_info = new_plugins



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc


class IpCommand(metaclass=abc.ABCMeta):

    TYPE_VETH = 'veth'
    TYPE_VLAN = 'vlan'
    TYPE_BRIDGE = 'bridge'

    @abc.abstractmethod
    def set(self, device, check_exit_code=None, state=None, mtu=None,
            address=None, promisc=None, master=None):
        """Method to set a parameter in an interface.

        :param   device: A network device (string)
        :param   check_exit_code: List of integers of allowed execution exit
                                  codes
        :param   state: String network device state
        :param   mtu: Integer MTU value
        :param   address: String MAC address
        :param   promisc: Boolean promiscuous mode
        :param   master: String the master device that this device belongs to
        :return: status of the command execution
        """

    @abc.abstractmethod
    def add(self, device, dev_type, check_exit_code=None, peer=None, link=None,
            vlan_id=None, ageing=None):
        """Method to add an interface.

        :param   device: A network device (string)
        :param   dev_type: String network device type (TYPE_VETH, TYPE_VLAN)
        :param   check_exit_code: List of integers of allowed execution exit
                                  codes
        :param   peer: String peer name, for veth interfaces
        :param   link: String root network interface name, 'device' will be a
                       VLAN tagged virtual interface
        :param   vlan_id: Integer VLAN ID for VLAN devices
        :param   ageing: integer value in seconds before learned
                         mac addresses are forgotten.
        :return: status of the command execution
        """

    @abc.abstractmethod
    def delete(self, device, check_exit_code=None):
        """Method to delete an interface.

        :param   device: A network device (string)
        :param   dev_type: String network device type (TYPE_VETH, TYPE_VLAN)
        :return: status of the command execution
        """

    @abc.abstractmethod
    def exists(self, device):
        """Method to dectect if a device exists.

        :param   device: A network device (string)
        :return: True if device exists else False
        """



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import warnings

import os_vif
from os_vif import objects
from os_vif.tests.unit import base


class TestVIFS(base.TestCase):

    def setUp(self):
        super(TestVIFS, self).setUp()

        os_vif.objects.register_all()

    def _test_vif(self, cls, **kwargs):
        vif = cls(**kwargs)

        prim = vif.obj_to_primitive()
        self.assertEqual("os_vif", prim["versioned_object.namespace"])
        vif2 = objects.vif.VIFBase.obj_from_primitive(prim)

        # The __eq__ function works by using obj_to_primitive()
        # and this includes a list of changed fields. Very
        # occassionally the ordering of the list of changes
        # varies, causing bogus equality failures. This is
        # arguably a bug in oslo.versionedobjects since the
        # set of changes fields should not affect equality
        # comparisons. Remove this hack once this is fixed:
        #
        # https://bugs.launchpad.net/oslo.versionedobjects/+bug/1563787
        vif.obj_reset_changes(recursive=True)
        vif2.obj_reset_changes(recursive=True)

        self.assertEqual(vif, vif2)

    def test_vif_generic(self):
        self._test_vif(objects.vif.VIFGeneric,
                       vif_name="vif123")

    def test_vif_bridge_plain(self):
        self._test_vif(objects.vif.VIFBridge,
                       vif_name="vif123",
                       bridge_name="br0")

    def test_port_profile_base_backport_1_0(self):
        datapath_offload = objects.vif.DatapathOffloadRepresentor(
            representor_name="felix",
            representor_address="0002:24:12.3")
        obj = objects.vif.VIFPortProfileBase(
            datapath_offload=datapath_offload)
        primitive = obj.obj_to_primitive(target_version='1.0')
        self.assertEqual('1.0', primitive['versioned_object.version'])
        data = primitive['versioned_object.data']
        self.assertNotIn('datapath_type', data)

    def test_vif_bridge_ovs(self):
        prof = objects.vif.VIFPortProfileOpenVSwitch(
            interface_id="07bd6cea-fb37-4594-b769-90fc51854ee9",
            profile_id="fishfood",
            datapath_type='netdev')
        self._test_vif(objects.vif.VIFOpenVSwitch,
                       vif_name="vif123",
                       bridge_name="br0",
                       port_profile=prof)

    def test_port_profile_ovs_backport_1_0(self):
        obj = objects.vif.VIFPortProfileOpenVSwitch(
            interface_id="07bd6cea-fb37-4594-b769-90fc51854ee9",
            profile_id="fishfood",
            datapath_type='netdev')
        primitive = obj.obj_to_primitive(target_version='1.0')
        self.assertEqual('1.0', primitive['versioned_object.version'])
        data = primitive['versioned_object.data']
        self.assertEqual('07bd6cea-fb37-4594-b769-90fc51854ee9',
                         data['interface_id'])
        self.assertEqual('fishfood', data['profile_id'])
        self.assertNotIn('datapath_type', data)

    def test_port_profile_ovs_backport_1_1(self):
        datapath_offload = objects.vif.DatapathOffloadRepresentor(
            representor_name="felix",
            representor_address="0002:24:12.3")
        obj = objects.vif.VIFPortProfileOpenVSwitch(
            interface_id="07bd6cea-fb37-4594-b769-90fc51854ee9",
            profile_id="fishfood",
            datapath_type='netdev',
            datapath_offload=datapath_offload)
        primitive = obj.obj_to_primitive(target_version='1.1')
        self.assertEqual('1.1', primitive['versioned_object.version'])
        data = primitive['versioned_object.data']
        self.assertEqual('07bd6cea-fb37-4594-b769-90fc51854ee9',
                         data['interface_id'])
        self.assertEqual('fishfood', data['profile_id'])
        self.assertEqual('netdev', data['datapath_type'])
        self.assertNotIn('datapath_offload', data)

    def test_port_profile_ovs_backport_1_2(self):
        obj = objects.vif.VIFPortProfileOpenVSwitch(
            interface_id="07bd6cea-fb37-4594-b769-90fc51854ee9",
            create_port=True)
        primitive = obj.obj_to_primitive(target_version='1.2')
        self.assertEqual('1.2', primitive['versioned_object.version'])
        data = primitive['versioned_object.data']
        self.assertEqual('07bd6cea-fb37-4594-b769-90fc51854ee9',
                         data['interface_id'])
        self.assertNotIn('create_port', data)

    def test_vif_direct_plain(self):
        self._test_vif(objects.vif.VIFDirect,
                       vif_name="vif123",
                       dev_address="0002:24:12.3")

    def test_vif_direct_vepa_qbg(self):
        prof = objects.vif.VIFPortProfile8021Qbg(
            manager_id=8,
            type_id=23,
            type_id_version=523,
            instance_id="72a00fee-2fbb-43e6-a592-c858d056fcfc")
        self._test_vif(objects.vif.VIFDirect,
                       vif_name="vif123",
                       port_profile=prof,
                       dev_address="0002:24:12.3")

    def test_vif_direct_vepa_qbh(self):
        prof = objects.vif.VIFPortProfile8021Qbh(
            profile_id="fishfood")
        self._test_vif(objects.vif.VIFDirect,
                       vif_name="vif123",
                       port_profile=prof,
                       dev_address="0002:24:12.3")

    def test_vif_vhost_user(self):
        self._test_vif(objects.vif.VIFVHostUser,
                       path="/some/socket.path",
                       mode=objects.fields.VIFVHostUserMode.CLIENT,
                       vif_name="vhu123")

    def test_vif_vhost_user_fp_ovs(self):
        prof = objects.vif.VIFPortProfileFPOpenVSwitch(
            interface_id="07bd6cea-fb37-4594-b769-90fc51854ee8",
            profile_id="fishfood",
            datapath_type='netdev',
            bridge_name="br-int",
            hybrid_plug=False)
        self._test_vif(objects.vif.VIFVHostUser,
                       path="/some/socket.path",
                       mode=objects.fields.VIFVHostUserMode.CLIENT,
                       vif_name="tap123",
                       port_profile=prof)

    def test_port_profile_fp_ovs_backport_1_0(self):
        obj = objects.vif.VIFPortProfileFPOpenVSwitch(
            interface_id="07bd6cea-fb37-4594-b769-90fc51854ee9",
            profile_id="fishfood",
            datapath_type='netdev',
            bridge_name="br-int",
            hybrid_plug=False)
        primitive = obj.obj_to_primitive(target_version='1.0')
        self.assertEqual('1.0', primitive['versioned_object.version'])
        data = primitive['versioned_object.data']
        self.assertEqual('07bd6cea-fb37-4594-b769-90fc51854ee9',
                         data['interface_id'])
        self.assertEqual('fishfood', data['profile_id'])
        self.assertEqual('br-int', data['bridge_name'])
        self.assertEqual(False, data['hybrid_plug'])
        self.assertNotIn('datapath_type', data)

    def test_port_profile_fp_ovs_backport_1_1(self):
        datapath_offload = objects.vif.DatapathOffloadRepresentor(
            representor_name="felix",
            representor_address="0002:24:12.3")
        obj = objects.vif.VIFPortProfileFPOpenVSwitch(
            interface_id="07bd6cea-fb37-4594-b769-90fc51854ee9",
            profile_id="fishfood",
            datapath_type='netdev',
            bridge_name="br-int",
            hybrid_plug=False,
            datapath_offload=datapath_offload)
        primitive = obj.obj_to_primitive(target_version='1.1')
        self.assertEqual('1.1', primitive['versioned_object.version'])
        data = primitive['versioned_object.data']
        self.assertEqual('07bd6cea-fb37-4594-b769-90fc51854ee9',
                         data['interface_id'])
        self.assertEqual('fishfood', data['profile_id'])
        self.assertEqual('br-int', data['bridge_name'])
        self.assertEqual(False, data['hybrid_plug'])
        self.assertEqual('netdev', data['datapath_type'])
        self.assertNotIn('datapath_offload', data)

    def test_vif_vhost_user_ovs_representor(self):
        prof = objects.vif.VIFPortProfileOVSRepresentor(
            interface_id="07bd6cea-fb37-4594-b769-90fc51854ee8",
            profile_id="fishfood",
            datapath_type='netdev',
            representor_name="tap123",
            representor_address="0002:24:12.3")
        self._test_vif(objects.vif.VIFVHostUser,
                       path="/some/socket.path",
                       mode=objects.fields.VIFVHostUserMode.CLIENT,
                       vif_name="tap123",
                       port_profile=prof)

    def test_port_profile_ovs_representor_backport_1_0(self):
        obj = objects.vif.VIFPortProfileOVSRepresentor(
            interface_id="07bd6cea-fb37-4594-b769-90fc51854ee9",
            profile_id="fishfood",
            datapath_type='netdev',
            representor_name="tap123",
            representor_address="0002:24:12.3")
        primitive = obj.obj_to_primitive(target_version='1.0')
        self.assertEqual('1.0', primitive['versioned_object.version'])
        data = primitive['versioned_object.data']
        self.assertEqual('07bd6cea-fb37-4594-b769-90fc51854ee9',
                         data['interface_id'])
        self.assertEqual('fishfood', data['profile_id'])
        self.assertEqual('tap123', data['representor_name'])
        self.assertEqual("0002:24:12.3", data['representor_address'])
        self.assertNotIn('datapath_type', data)

    def test_port_profile_ovs_representor_backport_1_1(self):
        datapath_offload = objects.vif.DatapathOffloadRepresentor(
            representor_name="felix",
            representor_address="0002:24:12.3")
        obj = objects.vif.VIFPortProfileOVSRepresentor(
            interface_id="07bd6cea-fb37-4594-b769-90fc51854ee9",
            profile_id="fishfood",
            datapath_type='netdev',
            representor_name="tap123",
            representor_address="0002:24:12.3",
            datapath_offload=datapath_offload)
        primitive = obj.obj_to_primitive(target_version='1.1')
        self.assertEqual('1.1', primitive['versioned_object.version'])
        data = primitive['versioned_object.data']
        self.assertEqual('07bd6cea-fb37-4594-b769-90fc51854ee9',
                         data['interface_id'])
        self.assertEqual('fishfood', data['profile_id'])
        self.assertEqual('tap123', data['representor_name'])
        self.assertEqual("0002:24:12.3", data['representor_address'])
        self.assertEqual('netdev', data['datapath_type'])
        self.assertNotIn('datapath_offload', data)

    def test_vif_vhost_user_generic_representor(self):
        datapath_offload = objects.vif.DatapathOffloadRepresentor(
            representor_name="felix",
            representor_address="0002:24:12.3")
        prof = objects.vif.VIFPortProfileBase(
            datapath_offload=datapath_offload,
            )
        self._test_vif(objects.vif.VIFVHostUser,
                       path="/some/socket.path",
                       mode=objects.fields.VIFVHostUserMode.SERVER,
                       vif_name="felix",
                       port_profile=prof)

    def test_vif_vhost_user_fp_lb(self):
        prof = objects.vif.VIFPortProfileFPBridge(bridge_name="brq456")
        self._test_vif(objects.vif.VIFVHostUser,
                       path="/some/socket.path",
                       mode=objects.fields.VIFVHostUserMode.CLIENT,
                       vif_name="tap123",
                       port_profile=prof)

    def test_vif_vhost_user_fp_tap(self):
        prof = objects.vif.VIFPortProfileFPTap(mac_address="fa:16:3e:4c:2c:30")
        self._test_vif(objects.vif.VIFVHostUser,
                       path="/some/socket.path",
                       mode=objects.fields.VIFVHostUserMode.CLIENT,
                       vif_name="tap123",
                       port_profile=prof)

    def test_vif_host_dev_plain(self):
        self._test_vif(
            objects.vif.VIFHostDevice,
            dev_type=objects.fields.VIFHostDeviceDevType.ETHERNET,
            dev_address="0002:24:12.3")

    def test_vif_host_dev_vepa_qbh(self):
        prof = objects.vif.VIFPortProfile8021Qbh(
            profile_id="fishfood")
        self._test_vif(objects.vif.VIFHostDevice,
                       dev_address="0002:24:12.3",
                       port_profile=prof)

    def test_vif_nested_dpdk_k8s(self):
        prof = objects.vif.VIFPortProfileK8sDPDK(
            l3_setup=False,
            selflink="/some/url",
            resourceversion="1")
        self._test_vif(
            objects.vif.VIFNestedDPDK,
            pci_adress="0002:24:12.3",
            dev_driver="virtio_pci",
            port_profile=prof)

    def test_port_profile_fp_bridge_backport_1_0(self):
        datapath_offload = objects.vif.DatapathOffloadRepresentor(
            representor_name="felix",
            representor_address="0002:24:12.3")
        obj = objects.vif.VIFPortProfileFPBridge(
            bridge_name='joe',
            datapath_offload=datapath_offload)
        primitive = obj.obj_to_primitive(target_version='1.0')
        self.assertEqual('1.0', primitive['versioned_object.version'])
        data = primitive['versioned_object.data']
        self.assertEqual('joe', data['bridge_name'])
        self.assertNotIn('datapath_type', data)

    def test_port_profile_fp_tap_backport_1_0(self):
        datapath_offload = objects.vif.DatapathOffloadRepresentor(
            representor_name="felix",
            representor_address="0002:24:12.3")
        obj = objects.vif.VIFPortProfileFPTap(
            mac_address='00:de:ad:be:ef:01',
            datapath_offload=datapath_offload)
        primitive = obj.obj_to_primitive(target_version='1.0')
        self.assertEqual('1.0', primitive['versioned_object.version'])
        data = primitive['versioned_object.data']
        self.assertEqual('00:de:ad:be:ef:01', data['mac_address'])
        self.assertNotIn('datapath_type', data)

    def test_port_profile_8021qbg_backport_1_0(self):
        datapath_offload = objects.vif.DatapathOffloadRepresentor(
            representor_name="felix",
            representor_address="0002:24:12.3")
        obj = objects.vif.VIFPortProfile8021Qbg(
            manager_id=42,
            type_id=43,
            type_id_version=44,
            instance_id='07bd6cea-fb37-4594-b769-90fc51854ee9',
            datapath_offload=datapath_offload)
        primitive = obj.obj_to_primitive(target_version='1.0')
        self.assertEqual('1.0', primitive['versioned_object.version'])
        data = primitive['versioned_object.data']
        self.assertEqual(42, data['manager_id'])
        self.assertEqual(43, data['type_id'])
        self.assertEqual(44, data['type_id_version'])
        self.assertEqual('07bd6cea-fb37-4594-b769-90fc51854ee9',
                         data['instance_id'])
        self.assertNotIn('datapath_type', data)

    def test_port_profile_8021qbh_backport_1_0(self):
        datapath_offload = objects.vif.DatapathOffloadRepresentor(
            representor_name="felix",
            representor_address="0002:24:12.3")
        obj = objects.vif.VIFPortProfile8021Qbh(
            profile_id='catfood',
            datapath_offload=datapath_offload)
        primitive = obj.obj_to_primitive(target_version='1.0')
        self.assertEqual('1.0', primitive['versioned_object.version'])
        data = primitive['versioned_object.data']
        self.assertEqual('catfood', data['profile_id'])
        self.assertNotIn('datapath_type', data)

    def test_port_profile_dpdk_k8s_backport_1_0(self):
        datapath_offload = objects.vif.DatapathOffloadRepresentor(
            representor_name="felix",
            representor_address="0002:24:12.3")
        obj = objects.vif.VIFPortProfileK8sDPDK(
            l3_setup=False,
            selflink="/some/url",
            resourceversion="1",
            datapath_offload=datapath_offload)
        primitive = obj.obj_to_primitive(target_version='1.0')
        self.assertEqual('1.0', primitive['versioned_object.version'])
        data = primitive['versioned_object.data']
        self.assertEqual(False, data['l3_setup'])
        self.assertEqual("/some/url", data['selflink'])
        self.assertEqual("1", data['resourceversion'])
        self.assertNotIn('datapath_type', data)

    def test_vif_host_dev_ovs_offload(self):
        datapath_offload = objects.vif.DatapathOffloadRepresentor(
            representor_name="felix",
            representor_address="0002:24:12.3")
        prof = objects.vif.VIFPortProfileOpenVSwitch(
            interface_id="07bd6cea-fb37-4594-b769-90fc51854ee8",
            profile_id="fishfood",
            datapath_type='netdev',
            datapath_offload=datapath_offload)
        self._test_vif(
            objects.vif.VIFHostDevice,
            dev_type=objects.fields.VIFHostDeviceDevType.ETHERNET,
            dev_address="0002:24:12.3",
            port_profile=prof)

    def test_pending_warnings_emitted_class_direct(self):
        with warnings.catch_warnings(record=True) as capture:
            warnings.simplefilter("always")
            pp = objects.vif.VIFPortProfileOVSRepresentor()
        self.assertEqual(1, len(capture))
        w = capture[0]
        self.assertEqual(PendingDeprecationWarning, w.category)
        self.assertEqual(pp.VERSION,
            objects.vif.VIFPortProfileOVSRepresentor.VERSION)
